
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/suspend.h>
#include "../suspend.h"
#include "debug.h"

/* -------------------- Display of PCP lists ---------------------- */
void show_pcp_lists(void)
{
	int cpu, temperature;
	struct zone *zone;

	for_each_zone(zone) {
		printk("%s per-cpu:", zone->name);

		if (!zone->present_pages) {
			printk(" empty\n");
			continue;
		} else
			printk("\n");

		for (cpu = 0; cpu < NR_CPUS; ++cpu) {
			struct per_cpu_pageset *pageset;

			if (!cpu_possible(cpu))
				continue;

			pageset = zone->pageset + cpu;

			for (temperature = 0; temperature < 2; temperature++)
				printk("cpu %d %s: low %d, high %d, batch %d, count %d.\n",
					cpu,
					temperature ? "cold" : "hot",
					pageset->pcp[temperature].low,
					pageset->pcp[temperature].high,
					pageset->pcp[temperature].batch,
					pageset->pcp[temperature].count);
		}
	}
}

/* -------------------- Free memory tracking ---------------------- */

int suspend_free_mem_values[MAX_FREEMEM_SLOTS][2];
/* These should match the enumerated type in suspend.h */
static char * suspend_free_mem_descns[MAX_FREEMEM_SLOTS] = {
	"Start/End      ", /* 0 */
	"Console Allocn ",
	"Drain pcp      ",
	"InUse map      ",
	"PS2 map        ",
	"Checksum map   ", /* 5 */
	"Reload pages   ",
	"Init plugins   ",
	"Memory pool    ",
	"Freezer        ",
	"Eat Memory     ", /* 10 */
	"Syncing        ",
	"               ",
	"Extent Pages   ",
	"Extra PD1 pages",
	"Writer storage ", /* 15 */
	"Header storage ",
	"Checksum pages ",
	"KStat data     ",
	"Debug Info     ",
	"Remove Image   ", /* 20 */
	"I/O            ",
	"I/O info       ",
	"Start one      ",
};

/* store_free_mem
 */

void suspend_store_free_mem(int slot, int side)
{
	static int last_free_mem;
	int this_free_mem = real_nr_free_pages();
	int i;

	BUG_ON(slot >= MAX_FREEMEM_SLOTS);

	suspend_message(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
			"Last free mem was %d. Is now %d. ",
			last_free_mem, this_free_mem);

	if (slot == 0) {
		if (!side)
			for (i = 1; i < MAX_FREEMEM_SLOTS; i++) {
				suspend_free_mem_values[i][0] = 0;
				suspend_free_mem_values[i][1] = 0;
			}
		suspend_free_mem_values[slot][side] = this_free_mem;
	} else
		suspend_free_mem_values[slot][side] += this_free_mem - last_free_mem;
	last_free_mem = this_free_mem;
	suspend_message(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
		"%s value %d now %d.\n",
		suspend_free_mem_descns[slot],
		side,
		suspend_free_mem_values[slot][side]);
}

/*
 * suspend_display_free_mem
 */
void suspend_display_free_mem(void)
{
	int i;


	if (!TEST_DEBUG_STATE(SUSPEND_MEMORY))
		return;
	
	suspend_message(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
		"Start:       %7d    End: %7d.\n",
		suspend_free_mem_values[0][0],
		suspend_free_mem_values[0][1]);
	
	for (i = 1; i < MAX_FREEMEM_SLOTS; i++)
		if (suspend_free_mem_values[i][0] + suspend_free_mem_values[i][1])
			suspend_message(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
				"%s %7d         %7d.\n",
				suspend_free_mem_descns[i],
				suspend_free_mem_values[i][0],
				suspend_free_mem_values[i][1]);
}

EXPORT_SYMBOL_GPL(suspend_store_free_mem);
