/*
 * kernel/power/device.c
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * Support for the driver model and ACPI sleep states.
 */

//#define DEBUG_DEVICE_TREE

#include <linux/device.h>
#include <linux/reboot.h>
#include <linux/suspend.h>
#include <linux/pm.h>
#include <linux/err.h>
#include "driver_model.h"
#include "suspend.h"

unsigned long suspend_powerdown_method = 5; /* S5 = off */
static int suspend_acpi_state_used = 0;

extern struct pm_ops * pm_ops;
static u32 pm_disk_mode_save;


#ifdef CONFIG_ACPI
extern u32 acpi_leave_sleep_state (u8 sleep_state);
#endif

void suspend_drivers_cleanup(void)
{
	if (pm_ops) pm_ops->pm_disk_mode = pm_disk_mode_save;
}

int suspend_drivers_init(void)
{
	if (pm_ops) {
		pm_disk_mode_save = pm_ops->pm_disk_mode;
		pm_ops->pm_disk_mode = PM_DISK_PLATFORM;
	}
			
	return 0;
}

static int suspend_enter_acpi_state(u32 state)
{
	int ret = 0;
	unsigned long flags;
	suspend_acpi_state_used = state;

	local_irq_save(flags);
	device_power_down(state);
	if (pm_ops && pm_ops->enter)
		ret = pm_ops->enter(state);
	else 
		printk("Failed to enter state.\n");

	device_power_up();
	local_irq_restore(flags);

	return ret;
}

static void suspend_power_off(void)
{
	machine_power_off();
	prepare_status(1, 0, "Probably not capable for powerdown.");
	while (1)
		cpu_relax();
	/* NOTREACHED */
}

int suspend_power_down_prepare(void)
{
	int ret = 0;
	
	if (suspend_powerdown_method == 3 ||
	    suspend_powerdown_method == 4)
		if (pm_ops && pm_ops->prepare)
			ret = pm_ops->prepare(suspend_powerdown_method);

	return ret;
}

int suspend_power_down_finish(void)
{
	int ret = 0;

	if (suspend_powerdown_method == 3 ||
			suspend_powerdown_method == 4)
		if (pm_ops && pm_ops->finish)
			ret = pm_ops->finish(suspend_powerdown_method);

	device_resume();

	return ret;
}

/*
 * suspend_power_down
 * Functionality   : Powers down or reboots the computer once the image
 *                   has been written to disk.
 * Key Assumptions : Able to reboot/power down via code called or that
 *                   the warning emitted if the calls fail will be visible
 *                   to the user (ie printk resumes devices).
 * Called From     : do_suspend2_suspend_2
 */

extern void apm_power_off(void);

void suspend_power_down(void)
{
	if (TEST_ACTION_STATE(SUSPEND_REBOOT)) {
		prepare_status(1, 0, "Ready to reboot.");
		device_shutdown();
		machine_restart(NULL);
	}

	if (suspend_powerdown_method >= 3 && suspend_powerdown_method <= 4) {
		prepare_status(1, 0, "Seeking to enter ACPI state");
		if (suspend_power_down_prepare()) {
			prepare_status(1, 0, "Preparing to enter ACPI state failed. Using normal powerdown.");
			goto abort_ACPI_sleep;
		}
		if (device_suspend(PM_SUSPEND_MEM)) {
			prepare_status(1, 0, "Suspending devices failed. Using normal powerdown.");
			suspend_power_down_finish();
			goto abort_ACPI_sleep;
		}
		if (suspend_enter_acpi_state(suspend_powerdown_method)) {
			prepare_status(1, 0, "Entering ACPI state failed. Using normal powerdown.");
			suspend_power_down_finish();
			goto abort_ACPI_sleep;
		}
		suspend_power_down_finish();
		return;
	} else
		prepare_status(1, 0, "Powering down.");
	
abort_ACPI_sleep:
	/* 
	 * FIXME: At resume, we'll still think we used S4 if we tried it.
	 * Does it matter?
	 */
	suspend_acpi_state_used = 0;
	device_shutdown();
	suspend_power_off();
}

void suspend_drivers_resume(int stage)
{
	switch (stage) {
		case SUSPEND_DRIVERS_IRQS_DISABLED:
			BUG_ON(!irqs_disabled());
			device_power_up();
			BUG_ON(!irqs_disabled());
			break;

		case SUSPEND_DRIVERS_IRQS_ENABLED:
			BUG_ON(irqs_disabled());
#ifdef CONFIG_ACPI
			if (suspend_acpi_state_used)
				acpi_leave_sleep_state(suspend_acpi_state_used);
#endif
			suspend_power_down_finish();
			BUG_ON(irqs_disabled());
			break;
	}
}

/*	suspend_drivers_suspend
 *	@stage - one of:
 *		1: Power down drivers not used in writing the image
 *		2: Quiesce drivers used in writing the image
 *		   (prior to making atomic copy)
 *		3: Power down drivers used in writing the image and
 *		   enter the suspend mode if configured to do so.
 *
 */
int suspend_drivers_suspend(int stage)
{
	int result = 0;
	const int state = PM_SUSPEND_DISK;

	switch (stage) {
		case SUSPEND_DRIVERS_IRQS_DISABLED:
			BUG_ON(!irqs_disabled());
			result = device_power_down(state);
			BUG_ON(!irqs_disabled());
			break;

		case SUSPEND_DRIVERS_IRQS_ENABLED:
			BUG_ON(irqs_disabled());
			result = device_suspend(state);
			BUG_ON(irqs_disabled());
			break;
	}
	return result;
}

