/*
 * kernel/power/extent.h
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * It contains declarations related to extents. Extents are
 * suspend's method of storing some of the metadata for the image.
 * See extent.c for more info.
 *
 */

struct extentchain {
	struct extent * first;
	struct extent * last;
	int size; /* size of the extent ie sum (max-min+1) */
	int allocs;
	int frees;
	int debug;
	int timesusedoptimisation;
	char * name;
	struct extent * lastaccessed, *prevtolastaccessed, *prevtoprev;
};

/*
 * We rely on extents not fitting evenly into a page.
 * The last four bytes are used to store the number
 * of the page, to make saving & reloading pages simpler.
 */
struct extent {
	unsigned long minimum;
	unsigned long maximum;
	struct extent * next;
};


#define EXTENTS_PER_PAGE (PAGE_SIZE / (sizeof(struct extent)))
#define EXTENTPAGELINK(x) ((unsigned long *) \
		((((unsigned long) x) & PAGE_MASK) + PAGE_SIZE - \
		 sizeof(unsigned long)))

#define extent_for_each(extentchain, extentpointer, value) \
if ((extentchain)->first) \
	for ((extentpointer) = (extentchain)->first, (value) = \
			(extentpointer)->minimum; \
	     ((extentpointer) && ((extentpointer)->next || (value) <= \
				 (extentpointer)->maximum)); \
	     (((value) == (extentpointer)->maximum) ? \
		((extentpointer) = (extentpointer)->next, (value) = \
		 ((extentpointer) ? (extentpointer)->minimum : 0)) : \
			(value)++))

/*
 * When using compression and expected_compression > 0,
 * we allocate fewer swap entries, so GET_EXTENT_NEXT can
 * validly run out of data to return.
 */
#define GET_EXTENT_NEXT(currentextent, currentval) \
{ \
	if (currentextent) { \
		if ((currentval) == (currentextent)->maximum) { \
			if ((currentextent)->next) { \
				(currentextent) = (currentextent)->next; \
				(currentval) = (currentextent)->minimum; \
			} else { \
				(currentextent) = NULL; \
				(currentval) = 0; \
			} \
		} else \
			currentval++; \
	} \
}

extern int max_extents_used;
extern int num_extent_pages;
int add_to_extent_chain(struct extentchain * chain, unsigned long value);
void put_extent_chain(struct extentchain * chain);
void print_chain(int debuglevel, struct extentchain * chain, int printasswap);
int free_extents(void);
int append_to_extent_chain(int chain, unsigned long min, unsigned long max);
void relativise_extents(void);
void relativise_chain(struct extentchain * chain);
void absolutise_extents(void);
void absolutise_chain(struct extentchain * chain);
unsigned long * get_extentpages_list_entry(int index);
int relocate_extentpages(void);

extern struct extent * first_extent_page, * last_extent_page;

#define EXTENT_RELATIVE(x) (struct extent *) ((((unsigned long) x) & \
			(PAGE_SIZE - 1)) | \
		((*EXTENTPAGELINK(x) & (PAGE_SIZE - 1)) << PAGE_SHIFT))
#define EXTENT_ABSOLUTE(entry) (struct extent *) \
	((((unsigned long) (entry)) & (PAGE_SIZE - 1)) | \
	 (unsigned long) get_extentpages_list_entry(((unsigned long) (entry)) >> PAGE_SHIFT))

/* swap_entry_to_extent_val & extent_val_to_swap_entry: 
 * We are putting offset in the low bits so consecutive swap entries
 * make consecutive extent values */
#define swap_entry_to_extent_val(swp_entry) (swp_entry.val)
#define extent_val_to_swap_entry(val) (swp_entry_t) { (val) }

