/*
 * kernel/power/suspend2_core/pageflags.c
 *
 * Copyright (C) 2004-2005 Nigel Cunningham <ncunningham@cyclades.com>
 * 
 * This file is released under the GPLv2.
 *
 * Routines for dynamically allocating and releasing bitmaps
 * used as pseudo-pageflags.
 *
 * Arrays are not contiguous. The first sizeof(void *) bytes are
 * the pointer to the next page in the bitmap. This allows us to
 * 1) work under low memory conditions where order 0 might be all
 *    that's available
 * 2) save the pages at suspend time, reload and relocate them as
 *    necessary at resume time without breaking anything (cf
 *    extent pages).
 */

#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/bitops.h>
#include <linux/list.h>
#include <linux/suspend.h>
#include "pageflags.h"
#include "../plugins.h"
#include "pagedir.h"

/* Maps used in copying the image back are in builtin.c */
dyn_pageflags_t in_use_map;
dyn_pageflags_t allocd_pages_map;
dyn_pageflags_t checksum_map;
#ifdef CONFIG_DEBUG_PAGEALLOC
dyn_pageflags_t unmap_map;
#endif

/* suspend_allocate_dyn_pageflags
 *
 * Description:	Allocate a bitmap for local page flags.
 * Arguments:	dyn_pageflags_t *:	Pointer to the bitmap.
 *
 * A version of allocate_dyn_pageflags that saves us then having
 * to relocate the flags when resuming.
 */
int suspend_allocate_dyn_pageflags(dyn_pageflags_t *pagemap)
{
	int i;

	BUG_ON(*pagemap);

	if (test_suspend_state(SUSPEND_NOW_RESUMING)) {
		/* 
		 * We use kfree unconditionally below. That's not a problem
		 * because we only use this path when reloading pageset1.
		 */
		*pagemap = (dyn_pageflags_t) get_nonconflicting_page();
		if (! *pagemap)
			return 1;
	} else 
		*pagemap = kmalloc(sizeof(void *) * PAGES_PER_BITMAP, GFP_ATOMIC);

	for (i = 0; i < PAGES_PER_BITMAP; i++) {
		(*pagemap)[i] = (unsigned long *) get_zeroed_page(GFP_ATOMIC);
		if (!(*pagemap)[i]) {
			printk("Error. Unable to allocate memory for "
					"local page flags.");
			free_dyn_pageflags(pagemap);
			return 1;
		}
	}
	return 0;
}

/* savepageflags
 *
 * Description: Save a set of pageflags.
 * Arguments:   dyn_pageflags_t *: Pointer to the bitmap being saved.
 */

void save_dyn_pageflags(dyn_pageflags_t pagemap)
{
	int i;

	if (!*pagemap)
		return;

	for (i = 0; i < PAGES_PER_BITMAP; i++)
		active_writer->ops.writer.write_header_chunk((char *) pagemap[i], PAGE_SIZE);
}

/* loadpageflags
 *
 * Description: Load a set of pageflags.
 * Arguments:   dyn_pageflags_t *: Pointer to the bitmap being loaded.
 *              (It must be allocated before calling this routine).
 */

void load_dyn_pageflags(dyn_pageflags_t pagemap)
{
	int i;

	if (!pagemap)
		return;

	for (i = 0; i < PAGES_PER_BITMAP; i++)
		active_writer->ops.writer.read_header_chunk((char *) pagemap[i], PAGE_SIZE);
}

void relocate_dyn_pageflags(dyn_pageflags_t *pagemap)
{
	int i;
	LIST_HEAD(rejected_pages);

	if (!*pagemap)
		return;

	relocate_page_if_required((void *) pagemap);

	for (i = 0; i < PAGES_PER_BITMAP; i++)
		relocate_page_if_required((void *) &((*pagemap)[i]));
}

/* Compare two sets of pageflags, treating map1 as the base.
 * We print how map2 differs.
 */

#define SHOW_DIFFS(a...) do { } while(0)
//#define SHOW_DIFFS(format, args...) printk(format, ## args)

int compare_pageflags(dyn_pageflags_t map1, dyn_pageflags_t map2)
{
	long map1_next_on = -1, map2_next_on = -1;
	long last_map1_on = -1, last_map2_on = -1;
	long diff_start;
	int diffs_found = 0;

	while (map1_next_on < (long) max_mapnr) {
		map1_next_on = __get_next_bit_on(map1, map1_next_on);
		map2_next_on = __get_next_bit_on(map2, map2_next_on);

		/* Handle case where map2 has bits on map1 doesn't */
		diff_start = map2_next_on;

		while (map2_next_on < map1_next_on) {
			last_map2_on = map2_next_on;
			map2_next_on = __get_next_bit_on(map2, map2_next_on);
			if (map2_next_on != last_map2_on + 1) {
				SHOW_DIFFS("Map2 adds pages %ld-%ld.\n", diff_start, last_map2_on);
				diff_start = map2_next_on;
				diffs_found = 1;
			}
		}

		if (diff_start < map1_next_on) {
			SHOW_DIFFS("Map2 adds pages %ld-%ld.\n", diff_start, map1_next_on - 1);
			diffs_found = 1;
		}

		/* Now handle where map2 doesn't have bits on that map1 does */
		diff_start = map1_next_on;

		while (map2_next_on > map1_next_on) {
			last_map1_on = map1_next_on;
			map1_next_on = __get_next_bit_on(map1, map1_next_on);
			if (map1_next_on != last_map1_on + 1) {
				SHOW_DIFFS("Map2 missing pages %ld-%ld.\n", diff_start, last_map1_on);
				diff_start = map1_next_on;
				diffs_found = 1;
			}
		}
	
		if (diff_start < map2_next_on) {
			SHOW_DIFFS("Map2 missing pages %ld-%ld.\n", diff_start, map1_next_on - 1);
			diffs_found = 1;
		}
	}
	
	return diffs_found;
}
