/*
 * kernel/power/pageflags.h
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * Suspend2 needs a few pageflags while working that aren't otherwise
 * used. To save the struct page pageflags, we dynamically allocate
 * a bitmap and use that. These are the only non order-0 allocations
 * we do.
 *
 * NOTE!!!
 * We assume that PAGE_SIZE - sizeof(void *) is a multiple of
 * sizeof(unsigned long). Is this ever false?
 */

#include <linux/dyn_pageflags.h>

extern dyn_pageflags_t in_use_map;
extern dyn_pageflags_t pageset1_map;
extern dyn_pageflags_t pageset1_copy_map;
extern dyn_pageflags_t pageset2_map;
extern dyn_pageflags_t allocd_pages_map;
extern dyn_pageflags_t checksum_map;
#ifdef CONFIG_DEBUG_PAGEALLOC
extern dyn_pageflags_t unmap_map;
#endif

/* 
 * inusemap is used in two ways: 
 * - During suspend, to tag pages which are not used (to speed up 
 *   count_data_pages);
 * - During resume, to tag pages which are in pagedir1. This does not tag 
 *   pagedir2 pages, so !== first use.
 */
#define PageInUse(page)	\
	test_bit(PAGEBIT(page), PAGE_UL_PTR(in_use_map, page))
#define SetPageInUse(page) \
	set_bit(PAGEBIT(page), PAGE_UL_PTR(in_use_map, page))
#define ClearPageInUse(page) \
	clear_bit(PAGEBIT(page), PAGE_UL_PTR(in_use_map, page))

#define PagePageset1(page) \
	(pageset1_map ? test_bit(PAGEBIT(page), PAGE_UL_PTR(pageset1_map, page)): 0)
#define SetPagePageset1(page) \
	set_bit(PAGEBIT(page), PAGE_UL_PTR(pageset1_map, page))
#define ClearPagePageset1(page) \
	clear_bit(PAGEBIT(page), PAGE_UL_PTR(pageset1_map, page))

#define PagePageset1Copy(page) \
	(pageset1_copy_map ? test_bit(PAGEBIT(page), PAGE_UL_PTR(pageset1_copy_map, page)): 0)
#define SetPagePageset1Copy(page) \
	set_bit(PAGEBIT(page), PAGE_UL_PTR(pageset1_copy_map, page))
#define ClearPagePageset1Copy(page) \
	clear_bit(PAGEBIT(page), PAGE_UL_PTR(pageset1_copy_map, page))

#define PagePageset2(page) \
	(pageset2_map ? test_bit(PAGEBIT(page), PAGE_UL_PTR(pageset2_map, page)): 0)
#define SetPagePageset2(page) \
	set_bit(PAGEBIT(page), PAGE_UL_PTR(pageset2_map, page))
#define ClearPagePageset2(page) \
	clear_bit(PAGEBIT(page), PAGE_UL_PTR(pageset2_map, page))

#define PageAllocd(page)	\
	test_bit(PAGEBIT(page), PAGE_UL_PTR(allocd_pages_map, page))
#define SetPageAllocd(page) \
	set_bit(PAGEBIT(page), PAGE_UL_PTR(allocd_pages_map, page))
#define ClearPageAllocd(page) \
	clear_bit(PAGEBIT(page), PAGE_UL_PTR(alloc_pages_map, page))

#define PageChecksumIgnore(page) \
	(checksum_map ? test_bit(PAGEBIT(page), PAGE_UL_PTR(checksum_map, page)) : 0)

#define SetPageChecksumIgnore(page) \
do { \
	if (checksum_map) \
		set_bit(PAGEBIT(page), PAGE_UL_PTR(checksum_map, page)); \
} while(0)

#define ClearPageChecksumIgnore(page) \
do { \
	if (checksum_map) \
		clear_bit(PAGEBIT(page), PAGE_UL_PTR(checksum_map, page)); \
} while(0)


#define SetPageUnmap(page) \
	set_bit(PAGEBIT(page), PAGE_UL_PTR(unmap_map, page))
#define PageUnmap(page) \
	test_bit(PAGEBIT(page), PAGE_UL_PTR(unmap_map, page))

#define __get_next_bit_on_safe(bitmap, counter) \
	do { \
		(counter)++; \
	} while(((counter) < max_mapnr) && \
		(!test_bit(PAGEBIT(pfn_to_page((counter))), \
			PAGE_UL_PTR(bitmap, pfn_to_page((counter))))));

static inline int __get_next_bit_on(dyn_pageflags_t bitmap, int counter)
{
	do {
		counter++;
	} while((counter < max_mapnr) &&
		(!test_bit(PAGEBIT(pfn_to_page(counter)),
			PAGE_UL_PTR(bitmap, pfn_to_page(counter)))));
	return counter;
}

#define BITMAP_FOR_EACH_SET(bitmap, counter) \
	for (counter = __get_next_bit_on(bitmap, -1); counter < max_mapnr; \
		counter = __get_next_bit_on(bitmap, counter))

extern void save_dyn_pageflags(dyn_pageflags_t pagemap);
extern void load_dyn_pageflags(dyn_pageflags_t pagemap);
int suspend_allocate_dyn_pageflags(dyn_pageflags_t *pagemap);
void relocate_dyn_pageflags(dyn_pageflags_t *pagemap);
int compare_pageflags(dyn_pageflags_t map1, dyn_pageflags_t map2);
