/*
 * /kernel/power/suspend2_core/proc.c
 */

#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/list.h>
#include <linux/suspend.h>
#include <asm/uaccess.h>

#include "suspend.h"
#include "../plugins.h"
#include "../block_io.h"
#include "../proc.h"

#define ALL_SETTINGS_VERSION 3

/*
 * suspend_write_compat_proc.
 *
 * This entry allows all of the settings to be set at once. 
 * It was originally for compatibility with pre- /proc/suspend
 * versions, but has been retained because it makes saving and
 * restoring the configuration simpler.
 */
static int suspend_write_compat_proc(struct file *file, const char * buffer,
		unsigned long count, void * data)
{
	char * buf1 = (char *) get_zeroed_page(GFP_ATOMIC), *curbuf, *lastbuf;
	char * buf2 = (char *) get_zeroed_page(GFP_ATOMIC); 
	int i, file_offset = 0, used_size = 0, reparse_resume_device = 0;
	unsigned long nextval;
	struct suspend_plugin_ops * plugin;
	struct plugin_header * plugin_header = NULL;

	if ((!buf1) || (!buf2))
		return -ENOMEM;

	while (file_offset < count) {
		int length = count - file_offset;
		if (length > (PAGE_SIZE - used_size))
			length = PAGE_SIZE - used_size;

		if (copy_from_user(buf1 + used_size, buffer + file_offset, length))
			return -EFAULT;

		curbuf = buf1;

		if (!file_offset) {
			/* Integers first */
			for (i = 0; i < 8; i++) {
				if (!*curbuf)
					break;
				lastbuf = curbuf;
				nextval = simple_strtoul(curbuf, &curbuf, 0);
				if (curbuf == lastbuf)
					break;
				switch (i) {
					case 0:
						if (nextval != ALL_SETTINGS_VERSION) {
							printk("Error loading saved settings. This data is for version %ld, but kernel module uses format %d.\n",
									nextval, ALL_SETTINGS_VERSION);
							goto out;
						}
					case 1:
						suspend_result = nextval;
						break;
					case 2:
						suspend_action = nextval;
						break;
					case 3:
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
						suspend_debug_state = nextval;
#endif
						break;
					case 4:
						suspend_default_console_level = nextval;
#ifndef CONFIG_SOFTWARE_SUSPEND_DEBUG
						if (suspend_default_console_level > 1)
							suspend_default_console_level = 1;
#endif
						break;
					case 5:
						image_size_limit = nextval;
						break;
					case 6:
#ifdef CONFIG_ACPI
						suspend_powerdown_method = nextval;
						if (suspend_powerdown_method < 3)
							suspend_powerdown_method = 3;
						if (suspend_powerdown_method > 5)
#endif
							suspend_powerdown_method = 5;
						break;
				}

				curbuf++;
				while (*curbuf == ' ')
					curbuf++;
			}

			if (count <= (curbuf - buf1))
				goto out;
			else {
				list_for_each_entry(plugin, &suspend_plugins, plugin_list)
					plugin->disabled = 1;
			}
		}

		if (((unsigned long) curbuf  & ~PAGE_MASK) + sizeof(plugin_header) > PAGE_SIZE)
			goto shift_buffer;
		
		/* Plugins */
		plugin_header = (struct plugin_header *) curbuf;

		if (((unsigned long) curbuf & ~PAGE_MASK) + sizeof(plugin_header) + plugin_header->data_length  > PAGE_SIZE)
			goto shift_buffer;
		
		if (plugin_header->magic != 0xADEDC0DE) {
			printk("Bad plugin data magic.\n");
			break;
		}

		plugin = find_plugin_given_name(plugin_header->name);

		if (plugin) {	/* May validly have config saved for a plugin not now loaded */
			if ((plugin->type == WRITER_PLUGIN) &&
			    ((!active_writer && plugin->disabled && !plugin_header->disabled) ||
			     (active_writer == plugin && plugin_header->disabled)))
				reparse_resume_device = 1;
			plugin->disabled = plugin_header->disabled;
			if (plugin_header->data_length)
				plugin->load_config_info(curbuf + sizeof(struct plugin_header), 
						plugin_header->data_length);
		} else
			printk("Data for plugin %s not used because not currently loaded.\n", plugin_header->name);

		curbuf += sizeof(struct plugin_header) + plugin_header->data_length;

shift_buffer:
		if (!(curbuf - buf1))
			break;

		file_offset += curbuf - buf1;
		
		used_size = PAGE_SIZE + buf1 - curbuf;
		memcpy(buf2, curbuf, used_size);
		memcpy(buf1, buf2, used_size);
	}
out:
	free_pages((unsigned long) buf1, 0);
	free_pages((unsigned long) buf2, 0);
	
	if (reparse_resume_device) {
		printk("Active writer disabled or no active writer and one or more just enabled. Reparsing resume device.\n");
		attempt_to_parse_resume_device();
	}
	
	return count;
}

/*
 * suspend_read_compat_proc.
 *
 * Like it's _write_ sibling, this entry allows all of the settings
 * to be read at once.
 * It too was originally for compatibility with pre- /proc/suspend
 * versions, but has been retained because it makes saving and
 * restoring the configuration simpler.
 */
static int suspend_read_compat_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	struct suspend_plugin_ops * this_plugin;
	char * buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	int index = 1, file_pos = 0, page_offset = 0, len;
	int copy_len = 0;
	struct plugin_header plugin_header;

	if (!buffer) {
		printk("Failed to allocate a buffer for getting "
				"plugin configuration info.\n");
		return -ENOMEM;
	}
		
	plugin_header.magic = 0xADEDC0DE;

	len = sprintf(buffer, "%d %ld %ld %ld %d %d %ld\n",
			ALL_SETTINGS_VERSION,
			suspend_result,
			suspend_action,
			suspend_debug_state,
			suspend_default_console_level,
			image_size_limit,
			suspend_powerdown_method);

	if (len >= off) {
		copy_len = (len < off + count) ? len - off : count - off;
		memcpy(page, buffer + off, copy_len);
		page_offset+= copy_len;
	}

	file_pos += len;

	/* 
	 * We have to know which data goes with which plugin, so we at
	 * least write a length of zero for a plugin. Note that we are
	 * also assuming every plugin's config data takes <= PAGE_SIZE.
	 */

	/* For each plugin (in registration order) */
	list_for_each_entry(this_plugin, &suspend_plugins, plugin_list) {

		/* Get the data from the plugin */
		if (this_plugin->save_config_info) {
			plugin_header.data_length = this_plugin->save_config_info(buffer);
		} else
			plugin_header.data_length = 0;

		if (file_pos > (off + count)) {
			file_pos += sizeof(struct plugin_header) + plugin_header.data_length;
			continue;
		}

		len = 0;
		if ((file_pos + sizeof(struct plugin_header) >= off) &&
		    (file_pos < (off + count))) {

			/* Save the details of the plugin */
			memcpy(plugin_header.name, this_plugin->name, 
					SUSPEND_MAX_PLUGIN_NAME_LENGTH);
			plugin_header.disabled = this_plugin->disabled;
			plugin_header.type = this_plugin->type;
			plugin_header.index = index++;
			
			copy_len = sizeof(struct plugin_header);

			if (copy_len + page_offset > count)
				copy_len = count - page_offset;

			memcpy(page + page_offset,
				 ((char *) &plugin_header) + off + page_offset - file_pos,
				 copy_len);	 

			page_offset += copy_len;
		}

		file_pos += sizeof(struct plugin_header);

		if (plugin_header.data_length && (file_pos >= off) && (file_pos < (off + count))) {
			copy_len = plugin_header.data_length;

			if (copy_len + page_offset > count + off)
				copy_len = count - page_offset;
			
			memcpy(page + page_offset,
				 buffer,
				 copy_len);	 

			page_offset += copy_len;

		}

		file_pos += plugin_header.data_length;
		
	}
	free_pages((unsigned long) buffer, 0);
	if (page_offset < count)
		*eof = 1;
	return page_offset;
}

static struct suspend_proc_data proc_params[] = {
	{ .filename			= "all_settings",
	  .permissions			= PROC_RW,
	  .type				= SUSPEND_PROC_DATA_CUSTOM,
	  .data = {
		  .special = {
	  		 .read_proc	= suspend_read_compat_proc,
			 .write_proc	= suspend_write_compat_proc,
		  }
	  }
	}
};

__init void suspend_core_proc_init(void)
{
	int i, numfiles = sizeof(proc_params) / sizeof(struct suspend_proc_data);

	for (i=0; i< numfiles; i++)
		suspend_register_procfile(&proc_params[i]);
}

void suspend_core_proc_exit(void)
{
	int i, numfiles = sizeof(proc_params) / sizeof(struct suspend_proc_data);

	for (i=0; i< numfiles; i++)
		suspend_unregister_procfile(&proc_params[i]);
}
