/*
 * kernel/power/suspend2.h
 *
 * Copyright (C) 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * This file is released under the GPLv2.
 *
 * It contains declarations used throughout swsusp and suspend2.
 *
 */
#ifndef KERNEL_POWER_SUSPEND_CORE_H
#define KERNEL_POWER_SUSPEND_CORE_H

#include <linux/delay.h>
#include "extent.h"
#include "pagedir.h"

/* ---------------------------- Suspend2 -------------------------------- */

/* Non-plugin data saved in our image header */
struct suspend_header {
	u32 version_code;
	unsigned long num_physpages;
	char machine[65];
	char version[65];
	int num_cpus;
	int page_size;
	unsigned long orig_mem_free;
	int num_extent_pages;
	struct extent * unused_extents;
	int pageset_2_size;
	int param0;
	int param1;
	int param2;
	int param3;
	int param4;
	int progress0;
	int progress1;
	int progress2;
	int progress3;
	int io_time[2][2];
	
	/* Implementation specific variables */
#ifdef KERNEL_POWER_SWSUSP_C
	suspend_pagedir_t *suspend_pagedir;
	unsigned int num_pbes;
#else
	struct pagedir pagedir;
#endif
};

extern void schedule_suspend_message(int message_number);
extern int suspend_min_free;

extern void save_avenrun_values(void);
extern void restore_avenrun_values(void);

#define SWAP_FILENAME_MAXLENGTH	32

extern int suspend_default_console_level;
extern int image_size_limit;

struct pageset_sizes_result {
	int size1; /* Can't be unsigned - breaks MAX function */
	int size1low;
	int size2;
	int size2low;
	int needmorespace;
};

#define KB(x) ((x) << (PAGE_SHIFT - 10))
#define MB(x) ((x) >> (20 - PAGE_SHIFT))

extern unsigned long get_max_low_pfn(void);

/*
 * XXX: We try to keep some more pages free so that I/O operations succeed
 * without paging. Might this be more?
 */
#define MIN_FREE_RAM (get_max_low_pfn() >> 7)

extern void prepare_status(int printalways, int clearbar, const char *fmt, ...);
extern void abort_suspend(const char *fmt, ...);

extern int suspend_snprintf(char * buffer, int buffer_size,
		const char *fmt, ...);

/* ------ console.c ------ */
void check_shift_keys(int pause, char * message);
unsigned long update_status(unsigned long value, unsigned long maximum,
		const char *fmt, ...);

extern int expected_compression_ratio(void);

#define MAIN_STORAGE_NEEDED(USE_ECR) \
	((pageset1_size + pageset2_size) * \
	 (USE_ECR ? expected_compression_ratio() : 100) / 100)

#define HEADER_BYTES_NEEDED \
	((num_extent_pages << PAGE_SHIFT) + \
	 sizeof(struct suspend_header) + \
	 sizeof(struct plugin_header) + \
	 (int) header_storage_for_plugins() + \
	 (PAGES_PER_BITMAP << PAGE_SHIFT) + \
	 num_plugins * \
	 	(sizeof(struct plugin_header) + sizeof(int)))
	
#define HEADER_STORAGE_NEEDED ((HEADER_BYTES_NEEDED + (int) PAGE_SIZE - 1) >> PAGE_SHIFT)

#define STORAGE_NEEDED(USE_ECR) \
	(MAIN_STORAGE_NEEDED(USE_ECR) + HEADER_STORAGE_NEEDED)

#define RAM_TO_SUSPEND (1 + max((pageset1_size - pageset2_sizelow), 0) + \
		MIN_FREE_RAM + memory_for_plugins())

extern int expected_compression_ratio(void);

extern unsigned int nr_suspends;

extern int suspend_wait_for_keypress(void);

#ifdef CONFIG_SMP
extern void smp_suspend(void);
extern void smp_continue(void);
#else
#define smp_suspend() do { } while(0)
#define smp_continue() do { } while(0)
#endif

extern struct pm_ops * pm_ops;
extern dev_t name_to_dev_t(char *line);
extern char _text[], _etext[], _edata[], __bss_start[], _end[];
extern void signal_wake_up(struct task_struct *t, int resume);

extern struct partial_device_tree * suspend_device_tree;


#ifdef CONFIG_DEBUG_PAGEALLOC
/* Returns whether it was already in the requested state */
extern int suspend_map_kernel_page(struct page * page, int enable);

extern void suspend_map_atomic_copy_pages(void);
extern void suspend_unmap_atomic_copy_pages(void);
#else
#define suspend_map_atomic_copy_pages() do { } while(0)
#define suspend_unmap_atomic_copy_pages() do { } while(0)
static inline int suspend_map_kernel_page(struct page * page, int enable)
{
	return 1;
}
#endif

extern unsigned long suspend_powerdown_method;
extern void attempt_to_parse_resume_device(void);

extern int real_nr_free_pages(void);

extern void suspend_check_mounts(void);

extern void mark_task_as_pageset1(struct task_struct *t);

extern void request_abort_suspend(void);

extern int suspend_get_keyboard_exclusive(void (* handler) (int));
extern void suspend_release_keyboard_exclusive(void);

#endif
