/*
 * Filewriter.c
 *
 * Copyright 2005 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * Distributed under GPLv2.
 * 
 * This file encapsulates functions for usage of a simple file as a
 * backing store. It is based upon the swapwriter, and shares the
 * same basic working. Here, though, we have nothing to do with
 * swapspace, and only one device to worry about.
 *
 * TODO LIST:
 * - Handling of block size
 * - Simple int for number of header pages
 * - (re)create block map extent lists
 * - support for expanding/shrinking file
 * - debugging info
 * - try to parse header device error handling
 * - signature handling
 */

#include <linux/suspend.h>
#include <linux/module.h>
#include <linux/blkdev.h>
#include <linux/file.h>
#include <linux/stat.h>
#include <linux/mount.h>

#include "suspend2_core/suspend.h"
#include "suspend2_core/debug.h"
#include "suspend2_core/version.h"
#include "block_io.h"
#include "proc.h"
#include "plugins.h"

#define SIGNATURE_VER 1

static struct suspend_proc_data filewriter_proc_data[];

/* --- Struct of pages stored on disk */

static struct suspend_plugin_ops filewriterops;

static struct inode * target_inode;
static struct block_device * target_bdev;
static int target_fd;

#if 0
// - Manage signature.
static int prepare_signature(struct submit_params * first_header_page,
		char * current_header);
#endif
static int parse_signature(char * signature, int restore);

extern asmlinkage long sys_open(const char __user * filename, int flags, int mode);
extern asmlinkage long sys_close(unsigned int fd);
#define translate_target (filewriter_proc_data[0].data.string.variable)
#if 0
// Higher Level
static int readahead_index = 0, readahead_submit_index = 0;
static int readahead_allocs = 0, readahead_frees = 0;

static char * filewriter_buffer = NULL;
static int filewriter_buffer_posn = 0;
static int filewriter_page_index = 0;
#endif

/*
 * ---------------------------------------------------------------
 *
 *     Internal Data Structures
 *
 * ---------------------------------------------------------------
 */

/* header_data contains data that is needed to reload pagedir1, and
 * is therefore saved in the suspend header.
 *
 * Pagedir2 data gets stored before pagedir1 (save order), and the first
 * page for pagedir1 to use is set when pagedir2 is written (when we know how
 * much storage it used). Since this first entry is almost certainly not at the
 * start of a extent, the firstoffset variable below tells us where to start in
 * the extent. All of this means we don't have to worry about getting different
 * compression ratios for the kernel and cache (when compressing the image).
 * We can simply allocate one pool of storage (size determined using expected
 * compression ratio) and use it without worrying whether one pageset
 * compresses better and the other worse (this is what happens). As long as the
 * user gets the expected compression right, it will work.
 */

struct {
	/* Extent chains for blocks */
	struct extentchain block_chain;
	
	/* Location of start of pagedir 1 */
	struct extent * pd1start_block_extent;
	unsigned long pd1start_block_offset;
	int pd1start_chain;

} header_data;


dev_t header_device = 0;
struct block_device * header_block_device = NULL;
struct extent * this_extent_page = NULL, * next_extent_page = NULL;
int headerblocksize = PAGE_SIZE;
int headerblock;

#if 0
/*
 * ---------------------------------------------------------------
 *
 *     Current state.
 *
 * ---------------------------------------------------------------
 */

/* Which pagedir are we saving/reloading? Needed so we can know whether to
 * remember the last block used at the end of writing pageset2, and
 * get that location when saving or reloading pageset1.*/
static int current_stream = 0;

/* Pointer to current entry being loaded/saved. */
static struct extent * currentblockextent = NULL;
static unsigned long currentblockoffset = 0;
static int currentblocksperpage = 0;

/* Header Page Information */
static int header_pages_allocated = 0;
static struct submit_params * first_header_submit_info = NULL,
 * last_header_submit_info = NULL, * current_header_submit_info = NULL;

/*
 * ---------------------------------------------------------------
 *
 *     User Specified Parameters
 *
 * ---------------------------------------------------------------
 */
#endif

static int resume_firstblock = 0;
static int resume_firstblocksize = PAGE_SIZE;

static dev_t resume_device = 0;
static struct block_device * resume_block_device = NULL;

/*
 * ---------------------------------------------------------------
 *
 *     Disk I/O routines
 *
 * ---------------------------------------------------------------
 */
char filewriter_target[256];

extern int expected_compression;

#define MARK_SWAP_SUSPEND 0
#define MARK_SWAP_RESUME 1

static int filewriter_invalidate_image(void);

#if 0
static int get_header_params(struct submit_params * headerpage)
{
	swp_entry_t entry = headerpage->swap_address;
	int swapfilenum = swp_type(entry);
	unsigned long offset = swp_offset(entry);
	struct swap_info_struct * sis = get_swap_info_struct(swapfilenum);
	sector_t sector = map_swap_page(sis, offset);

	headerpage->dev = sis->bdev,
	headerpage->blocks[0] = sector;
	headerpage->blocks_used = 1;
	headerpage->readahead_index = -1;
	return 0;
}
#endif

static inline int get_blocks_per_page(int chain)
{
	return 1;
#if 0
	int result = PAGE_SIZE /
		suspend_bio_ops.get_block_size(swap_info[chain].bdev);
	printk("Block size for chain %d is %d,\n",
			chain, result);
	return result;
#endif
}

static int try_to_parse_resume_device(char * commandline)
{
	resume_device = name_to_dev_t(commandline);
	if (!resume_device) {
		if (test_suspend_state(SUSPEND_BOOT_TIME))
			suspend_early_boot_message(1, SUSPEND_CONTINUE_REQ,
			"Failed to translate the device name into a "
			"device id.\n");
		else
			printk(name_suspend "Failed to translate \"%s\""
				" into a device id.\n", commandline);
		return 1;
	}
	
	resume_block_device = open_by_devnum(resume_device, FMODE_READ);

	if (IS_ERR(resume_block_device)) {
		printk("Open by devnum returned %p given %x.\n",
				resume_block_device, resume_device);
		if (test_suspend_state(SUSPEND_BOOT_TIME))
			suspend_early_boot_message(1, SUSPEND_CONTINUE_REQ,
				"Failed to get access to the device on which"
				" Software Suspend's header should be found.");
		else
			printk("Failed to get access to the device on which "
				"Software Suspend's header should be found.\n");
		return 1;
	}

	return 0;
}

#if 0
static int try_to_parse_header_device(void)
{
	header_block_device = open_by_devnum(header_device, FMODE_READ);

	if (IS_ERR(header_block_device)) {
		if (suspend_early_boot_message(1,SUSPEND_CONTINUE_REQ,  
				"Failed to get access to the "
				"resume header device.\nYou could be "
				"booting with a 2.6 kernel when you "
				"suspended a 2.4 kernel."))
			filewriter_invalidate_image();

		return -EINVAL;
	}

	if (set_blocksize(header_block_device, PAGE_SIZE) < 0) {
		if (suspend_early_boot_message(1, SUSPEND_CONTINUE_REQ,
			"Failed to set the blocksize for our storage device."))
				do { } while(0);
		filewriter_invalidate_image();
		return -EINVAL;
	}

	return 0;
}
#endif

static void filewriter_noresume_reset(void)
{
 	/* 
	 * If we have read part of the image, we might have filled header_data with
	 * data that should be zeroed out.
	 */

	memset((char *) &header_data, 0, sizeof(header_data));
}

/*
 *
 */

int parse_signature(char * header, int restore)
{
	int have_image = !memcmp("HaveImage\n",header,10);
	int non_image_header = !memcmp("Suspend2\n", header, 9);

	printk("Imageheader is %s.\n", header);

	if (!have_image && !non_image_header)
		return -1;

	if (non_image_header)
		return 0;
	
	clear_suspend_state(SUSPEND_RESUMED_BEFORE);

	if (header[11] & 1)
		set_suspend_state(SUSPEND_RESUMED_BEFORE);

	/* Invalidate Image */
	if (restore)
		strcpy(header, "Suspend2\n");

	return 1;
}

#if 0
/*
 * prepare_signature
 */

static int prepare_signature(struct submit_params * header_page_info,
		char * current_header)
{
	/* 
	 * Explicitly put the \0 that clears the 'tried to resume from
	 * this image before' flag.
	 */
	strcpy(current_header, "HaveImage\n\0");

	return 0;
}
#endif

extern int signature_check(char * header, int fix);

#if 0
static int free_swap_pages_for_header(void)
{
	if (!first_header_submit_info)
		return 1;

	PRINTFREEMEM("at start of free_swap_pages_for_header");

	while (first_header_submit_info) {
		struct submit_params * next = first_header_submit_info->next;
		if (first_header_submit_info->swap_address.val)
			swap_free(first_header_submit_info->swap_address);
		kfree(first_header_submit_info);
		first_header_submit_info = next;
	}
	
	suspend_message(SUSPEND_WRITER, SUSPEND_LOW, 1,
			" Freed %d swap pages in free_swap_pages_for_header.\n",
			header_pages_allocated);
	first_header_submit_info = last_header_submit_info = NULL;
	header_pages_allocated = 0;
	PRINTFREEMEM("at end of free_swap_pages_for_header");
	suspend_store_free_mem(SUSPEND_FREE_HEADER_STORAGE, 1);
	return 0;
}
#endif

#if 0
static void get_main_pool_phys_params(void)
{
	struct extent * extentpointer = NULL;
	unsigned long address;
	int i;
	
	for (i = 0; i < MAX_SWAPFILES; i++)
		if (header_data.block_chain[i].first)
			put_extent_chain(&header_data.block_chain[i]);

	extent_for_each(&header_data.swapextents, extentpointer, address)
		get_phys_params(extent_val_to_swap_entry(address));
	
	sector_t sector = map_swap_page(sis, offset);
	add_to_extent_chain(&header_data.block_chain[swapfilenum], sector);
}
#endif

extern void put_extent(struct extent * extent);

static unsigned long filewriter_storage_allocated(void)
{
#if 0
	return (header_data.swapextents.size + header_pages_allocated);
#endif
	return 0;
}

static long filewriter_storage_available(void)
{
#if 0
	si_swapinfo(&swapinfo);
	return (swapinfo.freeswap + (long) filewriter_storage_allocated());
#endif
	return 0;
}

static int filewriter_initialise(void)
{
	target_fd = sys_open(translate_target, O_RDWR, 0);
	if (target_fd < 0)
		return target_fd;

	target_inode = current->files->fd[target_fd]->f_dentry->d_inode;
	target_bdev = target_inode->i_bdev ? target_inode->i_bdev : target_inode->i_sb->s_bdev;

	return 0;
}

static void filewriter_cleanup(void)
{
	sys_close(target_fd);
}

static int filewriter_release_storage(void)
{
	if ((TEST_ACTION_STATE(SUSPEND_KEEP_IMAGE)) && test_suspend_state(SUSPEND_NOW_RESUMING))
		return 0;

	/* Free metadata */

	/* Resize the file back down to 1 block */
	target_inode->i_blocks = 1 << (target_inode->i_sb->s_blocksize_bits - 9);
	target_inode->i_op->truncate(target_inode);
	return 0;
}

static long filewriter_allocate_header_space(unsigned long space_really_requested)
{
#if 0
	/* space_requested was going to be in bytes... not yet */
	int i, ret = 0;
	unsigned long space_requested;

	/* 
	 * Up to here in the process, we haven't taken account of the fact
	 * that we need an extra four bytes per 4092 bytes written for link
	 * to the next page on which the header will be written. We factor
	 * that in here.
	 */
	space_requested = ((4096 * space_really_requested + 4091) / 4092);
	space_requested = (space_requested * 4 + 4091) / 4092;
	space_requested += space_really_requested;
	
	PRINTFREEMEM("at start of allocate_header_space");
	
	for (i=(header_pages_allocated+1); i<=space_requested; i++) {
		struct submit_params * new_submit_param;
		
		/* Get a submit structure */
		new_submit_param = kmalloc(sizeof(struct submit_params), GFP_ATOMIC);
		
		if (!new_submit_param) {
			header_pages_allocated = i - 1;
			printk("Failed to kmalloc a struct submit param.\n");
			ret = -ENOMEM;
			goto out;
		}

		memset(new_submit_param, 0, sizeof(struct submit_params));

		if (last_header_submit_info) {
			last_header_submit_info->next = new_submit_param;
			last_header_submit_info = new_submit_param;
		} else
			last_header_submit_info = first_header_submit_info =
				new_submit_param;

		/* Get swap entry */
		new_submit_param->swap_address = get_swap_page();
		
		if ((!new_submit_param->swap_address.val) &&
			       (header_data.swapextents.first)) {
			/*
			 *  Steal one from pageset swap chain. If, as a result,
			 *  it is too small, more swap will be allocated or
			 *  memory eaten.
			 */

			new_submit_param->swap_address =
				extent_val_to_swap_entry(
					header_data.swapextents.first->minimum);
			if (header_data.swapextents.first->minimum <
					header_data.swapextents.first->maximum)
				header_data.swapextents.first->minimum++;
			else {
				struct extent * oldfirst =
					header_data.swapextents.first;
				header_data.swapextents.first = oldfirst->next;
				header_data.swapextents.frees++;
				header_data.swapextents.prevtoprev =
				 header_data.swapextents.prevtolastaccessed =
				 header_data.swapextents.lastaccessed = NULL;
				if (header_data.swapextents.last == oldfirst)
					header_data.swapextents.last = NULL;
				put_extent(oldfirst);
			}
			
			header_data.swapextents.size--;

			/*
			 * Recalculate block chains for main pool.
			 * We don't assume blocks are at start of a chain and
			 * don't know how many blocks per swap entry.
			 */
			get_main_pool_phys_params();
		}
		if (!new_submit_param->swap_address.val) {
			free_swap_pages_for_header();
			printk("Unable to allocate swap page for header.\n");
			ret = -ENOMEM;
			goto out;
		}
		if (get_header_params(new_submit_param)) {
			printk("Failed to get header parameters.\n");
			ret = -EFAULT;
			goto out;
		}
		suspend_message(SUSPEND_HEADER, SUSPEND_MEDIUM, 0,
			" Got header page %d/%d. Dev is %x. Block is %lu. "
			"Blocksperpage is %d.\n",
			i, space_requested,
			new_submit_param->dev,
			new_submit_param->blocks[0],
			new_submit_param->blocks_used);
	}
	header_pages_allocated = space_requested;
	suspend_message(SUSPEND_HEADER, SUSPEND_LOW, 1,
			" Have %d swap pages in filewriter::"
			"allocate_header_space.\n",
			header_pages_allocated);
out:
	PRINTFREEMEM("at end of filewriter::allocate_header_space");
	suspend_store_free_mem(SUSPEND_FREE_HEADER_STORAGE, 0);
	return ret;
#endif
	return 0;
}

static int filewriter_allocate_storage(unsigned long space_requested)
{
#if 0
	int i, swapcount = 0, result = 0;
	int lastsize = header_data.swapextents.size;
	int numwanted = (int) (space_requested);
	int pages_to_get = numwanted - header_data.swapextents.size;
	
	if (numwanted < 1)
		return 0;

	suspend_message(SUSPEND_WRITER, SUSPEND_MEDIUM, 0,
		"Started with swapextents.size == %d. "
		"Seeking to allocate %d more.\n",
		header_data.swapextents.size,
		pages_to_get);

	for(i=0; i < pages_to_get; i++) {
		swp_entry_t entry;
		suspend_message(SUSPEND_WRITER, SUSPEND_VERBOSE, 1, "");
		entry = get_swap_page();
		if (!entry.val) {
			suspend_message(SUSPEND_WRITER, SUSPEND_MEDIUM, 0,
				"Allocated %d/%d swap pages for main pool "
				"in allocate_swap.\n",
				swapcount, numwanted);
			result = -ENOSPC;
			goto out;
		}
		swapcount++;
		{
			int result =
				add_to_extent_chain(&header_data.swapextents,
					       swap_entry_to_extent_val(entry));
			if (result)
				printk("add_to_extent_chain returned %d.\n",
						result);
		}
		if (header_data.swapextents.size != (lastsize + 1))
			printk("swapextents.size == %d.\n",
					header_data.swapextents.size);
		lastsize = header_data.swapextents.size;
		check_shift_keys(0, NULL);
		if (TEST_RESULT_STATE(SUSPEND_ABORTED))
			break;
	}
	suspend_message(SUSPEND_WRITER, SUSPEND_MEDIUM, 0,
		" Allocated %d/%d swap pages in allocate_swap.\n",
		swapcount, numwanted);
	
	suspend_message(SUSPEND_WRITER, SUSPEND_MEDIUM, 0,
		"Finished with swapextents.size == %d.\n",
		header_data.swapextents.size);

out:
	get_main_pool_phys_params();
	
	/* Any memory we allocate will be for extent pages */
	suspend_store_free_mem(SUSPEND_FREE_EXTENT_PAGES, 0);
	return result;
#endif
	return 0;
}

static int filewriter_write_header_chunk(char * buffer, int buffer_size);
//static int header_bytes_written;

static int filewriter_write_header_init(void)
{
#if 0
	int i;

	header_bytes_written = 0;

	for (i = 0; i < MAX_SWAPFILES; i++)
		if (swap_info[i].swap_file) {
			header_data.swapdevs[i] = swap_info[i].bdev->bd_dev;
			header_data.blocksizes[i] =
				block_size(swap_info[i].bdev);
		}

	filewriter_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	header_link =
		(unsigned long *) (filewriter_buffer + BYTES_PER_HEADER_PAGE);
	filewriter_page_index = 1;

	current_header_submit_info = first_header_submit_info;
	
	/* Info needed to bootstrap goes at the start of the header.
	 * First we save the 'header_data' struct, including the number
	 * of header pages. Then we save the structs containing data needed
	 * for reading the header pages back.
	 * Note that even if header pages take more than one page, when we
	 * read back the info, we will have restored the location of the
	 * next header page by the time we go to use it.
	 */
	filewriter_write_header_chunk((char *) &header_data, 
			sizeof(header_data));

#endif
	return 0;
}

static int filewriter_write_header_chunk(char * buffer, int buffer_size)
{
#if 0
	int bytes_left = buffer_size;
	
	/* 
	 * We buffer the writes until a page is full and to use the last
	 * sizeof(swp_entry_t) bytes for links between pages. This is 
	 * totally transparent to the caller.
	 *
	 * Note also that buffer_size can be > PAGE_SIZE.
	 */

	header_bytes_written += buffer_size;
	
	suspend_message(SUSPEND_HEADER, SUSPEND_HIGH, 0,
		"\nStart of write_header_chunk loop with %d bytes to store.\n",
		buffer_size);

	while (bytes_left) {
		char * source_start = buffer + buffer_size - bytes_left;
		char * dest_start = filewriter_buffer + swapwriter_buffer_posn;
		int dest_capacity = BYTES_PER_HEADER_PAGE - filewriter_buffer_posn;
		swp_entry_t next_header_page;
		if (bytes_left <= dest_capacity) {
			suspend_message(SUSPEND_HEADER, SUSPEND_HIGH, 0,
				"Storing %d bytes from %p-%p in page %d, %p-%p.\n",
				bytes_left,
				source_start, source_start + bytes_left - 1,
				filewriter_page_index,
				dest_start, dest_start + bytes_left - 1);
			memcpy(dest_start, source_start, bytes_left);
			filewriter_buffer_posn += bytes_left;
			return 0;
		}
	
		/* A page is full */
		suspend_message(SUSPEND_HEADER, SUSPEND_HIGH, 0,
			"Storing %d bytes from %p-%p in page %d, %p-%p.\n",
			dest_capacity,
			source_start, source_start + dest_capacity - 1,
			filewriter_page_index,
			dest_start, dest_start + dest_capacity - 1);
		memcpy(dest_start, source_start, dest_capacity);
		bytes_left -= dest_capacity;

		BUG_ON(!current_header_submit_info);

		if (!current_header_submit_info->next) {
			suspend_message(SUSPEND_HEADER, SUSPEND_HIGH, 0,
				"This submit_info is the last one. Link zeroed.\n");
			*header_link = 0;
		} else {
			next_header_page =
				swp_entry(swp_type(
				current_header_submit_info->next->swap_address),
				current_header_submit_info->next->blocks[0]);

			*header_link = next_header_page.val;
		
			suspend_message(SUSPEND_HEADER, SUSPEND_HIGH, 0,
				"Header link is at %p. "
				"Contents set to swap device #%ld, block %ld.\n",
				header_link,
				(long) swp_type(next_header_page),
				swp_offset(next_header_page));
		}

		suspend_message(SUSPEND_HEADER, SUSPEND_HIGH, 0,
			"Writing header page %d/%d. "
			"Dev is %x. Block is %lu. Blocksperpage is %d.\n",
			filewriter_page_index, header_pages_allocated,
			current_header_submit_info->dev->bd_dev,
			current_header_submit_info->blocks[0],
			current_header_submit_info->blocks_used);
		
		current_header_submit_info->page =
			virt_to_page(filewriter_buffer);
		check_shift_keys(0, NULL);
		suspend_bio_ops.submit_io(WRITE, current_header_submit_info, 0);

		filewriter_buffer_posn = 0;
		filewriter_page_index++;
		current_header_submit_info = current_header_submit_info->next;
	}

#endif
	return 0;
}

static int filewriter_write_header_cleanup(void)
{
#if 0
	/* Write any unsaved data */
	if (filewriter_buffer_posn) {
		*header_link = 0;

		suspend_message(SUSPEND_HEADER, SUSPEND_HIGH, 0,
			"Writing header page %d/%d. "
			"Dev is %x. Block is %lu. Blocksperpage is %d.\n",
			filewriter_page_index, header_pages_allocated,
			current_header_submit_info->dev->bd_dev,
			current_header_submit_info->blocks[0],
			current_header_submit_info->blocks_used);
		
		current_header_submit_info->page =
			virt_to_page(filewriter_buffer);
		suspend_bio_ops.submit_io(WRITE, 
				current_header_submit_info, 0);
	}

	/* Adjust swap header */
	suspend_bio_ops.bdev_page_io(READ, resume_block_device, resume_firstblock,
			virt_to_page(filewriter_buffer));

	prepare_signature(first_header_submit_info,
		((union swap_header *) filewriter_buffer)->magic.magic);
		
	suspend_bio_ops.bdev_page_io(WRITE, resume_block_device, resume_firstblock,
			virt_to_page(filewriter_buffer));

	free_pages((unsigned long) filewriter_buffer, 0);
	filewriter_buffer = NULL;
	header_link = NULL;
	
	suspend_bio_ops.finish_all_io();

#endif
	return 0;
}

/* ------------------------- HEADER READING ------------------------- */

/*
 * read_header_init()
 * 
 * Description:
 * 1. Attempt to read the device specified with resume2=.
 * 2. Check the contents of the swap header for our signature.
 * 3. Warn, ignore, reset and/or continue as appropriate.
 * 4. If continuing, read the filewriter configuration section
 *    of the header and set up block device info so we can read
 *    the rest of the header & image.
 *
 * Returns:
 * May not return if user choose to reboot at a warning.
 * -EINVAL if cannot resume at this time. Booting should continue
 * normally.
 */

static int filewriter_read_header_init(void)
{
#if 0
	int i;
	
	filewriter_page_index = 1;

	filewriter_buffer = (char *) get_zeroed_page(GFP_ATOMIC);

	if (!header_device) {
		printk("read_header_init called when we haven't "
				"verified there is an image!\n");
		return -EINVAL;
	}

	/* 
	 * If the header is not on the resume_device, get the resume device first.
	 */
	if (header_device != resume_device) {
		int result = try_to_parse_header_device();

		if (result)
			return result;
	} else
		header_block_device = resume_block_device;

	/* Read filewriter configuration */
	suspend_bio_ops.bdev_page_io(READ, header_block_device, headerblock,
			virt_to_page((unsigned long) filewriter_buffer));
	//FIXME Remember location of next page to be read.
	
	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
		"Retrieving %d bytes from %x:%x to page %d, %p-%p.\n",
		header_block_device->bd_dev, headerblock,
		sizeof(header_data),
		filewriter_page_index,
		filewriter_buffer, swapwriter_buffer + sizeof(header_data) - 1);
	memcpy(&header_data, filewriter_buffer, sizeof(header_data));
	
	/* Restore device info */
	for (i = 0; i < MAX_SWAPFILES; i++) {
		dev_t thisdevice = header_data.swapdevs[i];
		
		suspend_message(SUSPEND_WRITER, SUSPEND_VERBOSE, 1,
				"Swap device %d is %x.", i, thisdevice);
		
		if (!thisdevice)
			continue;

		if (thisdevice == resume_device) {
			suspend_message(SUSPEND_WRITER, SUSPEND_VERBOSE, 0,
					"Resume root device %x", thisdevice);
			swap_info[i].bdev = resume_block_device;
			/* Mark as used so the device doesn't get suspended. */
			swap_info[i].swap_file = (struct file *) 0xffffff;
			continue;
		}

		if (thisdevice == header_device) {
			suspend_message(SUSPEND_WRITER, SUSPEND_VERBOSE, 0,
					"Resume header device %x", thisdevice);
			swap_info[i].bdev = header_block_device;
			/* Mark as used so the device doesn't get suspended. */
			swap_info[i].swap_file = (struct file *) 0xffffff;
			continue;
		}

		open_other_swap_device(i, thisdevice);
		swap_info[i].swap_file = (struct file *) 0xffffff;
	}

	filewriter_buffer_posn = sizeof(header_data);

#endif
	return 0;
}

static int filewriter_read_header_chunk(char * buffer, int buffer_size)
{
#if 0
	int bytes_left = buffer_size, ret = 0;
	
	check_shift_keys(0, "");

	/* Read a chunk of the header */
	while ((bytes_left) && (!ret)) {
		swp_entry_t next =
		   ((union p_diskpage) filewriter_buffer).pointer->link.next;
		struct block_device * dev = swap_info[swp_type(next)].bdev;
		int pos = swp_offset(next);
		char * dest_start = buffer + buffer_size - bytes_left;
		char * source_start =
			filewriter_buffer + swapwriter_buffer_posn;
		int source_capacity =
			BYTES_PER_HEADER_PAGE - filewriter_buffer_posn;

		if (bytes_left <= source_capacity) {
			suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
				"Retrieving %d bytes from page %d, "
				"%p-%p into %p-%p.\n",
				bytes_left,
				filewriter_page_index,
				source_start, source_start + bytes_left - 1,
				dest_start, dest_start + bytes_left - 1);
			memcpy(dest_start, source_start, bytes_left);
			filewriter_buffer_posn += bytes_left;
			return buffer_size;
		}

		/* Next to read the next page */
		suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
			"Retrieving %d bytes from page %d, %p-%p to %p-%p.\n",
			source_capacity,
			filewriter_page_index,
			source_start, source_start + source_capacity - 1,
			dest_start, dest_start + source_capacity - 1);
		memcpy(dest_start, source_start, source_capacity);
		bytes_left -= source_capacity;

		suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
		 "Header link is at %p. Contents set to %lx = "
		 "swap device #%x, block %d.\n",
		 &((union p_diskpage) filewriter_buffer).pointer->link.next,
		 ((union p_diskpage) filewriter_buffer).pointer->link.next.val,
		 dev->bd_dev, pos);

		filewriter_page_index++;

		suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
			"Reading header page %d. Dev is %x. Block is %lu.\n",
			filewriter_page_index, dev->bd_dev, pos);
		
		suspend_bio_ops.bdev_page_io(READ, dev, pos, virt_to_page(filewriter_buffer));

		filewriter_buffer_posn = 0;
	}

	return buffer_size - bytes_left;
#endif
	return 0;
}

static int filewriter_read_header_cleanup(void)
{
#if 0
	free_pages((unsigned long) filewriter_buffer, 0);
#endif
	return 0;
}

static int filewriter_prepare_save_extents(void)
{
#if 0
	int i;

	relativise_chain(&header_data.swapextents);
	
	for (i = 0; i < MAX_SWAPFILES; i++)
		relativise_chain(&header_data.block_chain[i]);
	
	header_data.pd1start_block_extent =
		EXTENT_RELATIVE(header_data.pd1start_block_extent);

	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
		"Pagedir1 firstblockextent is %p.\n",
		header_data.pd1start_block_extent);

#endif
	return 0;
}

static int filewriter_post_load_extents(void)
{
#if 0
	int i;
	
	absolutise_chain(&header_data.swapextents);

	for (i = 0; i < MAX_SWAPFILES; i++)
		absolutise_chain(&header_data.block_chain[i]);

	header_data.pd1start_block_extent =
		EXTENT_ABSOLUTE(header_data.pd1start_block_extent);
	
#endif
	return 0;
}

static int filewriter_write_init(int stream_number)
{
#if 0
	if (stream_number == 1) {
		currentblockextent = header_data.pd1start_block_extent;
		currentblockoffset = header_data.pd1start_block_offset;
	} else {
		currentblockextent = header_data.block_chain.first;
		currentblockoffset = currentblockextent->minimum;
	}

	BUG_ON(!currentblockextent);

	currentblocksperpage = PAGE_SIZE / 
		suspend_bio_ops.get_block_size(swap_info[currentblockchain].bdev);

	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
		"Stream %d beginning from position: chain %d. "
		"extent %p, block %ld.\n",
		stream_number,
		currentblockchain, currentblockextent, currentblockoffset);
	
	filewriter_page_index = 1;
	current_stream = stream_number;

	suspend_bio_ops.reset_io_stats();

#endif
	return 0;
}

static int filewriter_write_chunk(struct page * buffer_page)
{
#if 0
	int i;
	struct submit_params submit_params;

	if (TEST_ACTION_STATE(SUSPEND_TEST_FILTER_SPEED))
		return 0;
		
	if (currentblockchain == MAX_SWAPFILES) {
		printk("Error! We have run out of blocks for writing data.\n");
		for (i = 0; i < MAX_SWAPFILES; i++) {
			if (!swap_info[i].swap_file)
				printk("Swap slot %d is unused.\n", i);
			else
				printk("Swap slot %d is device %x.\n",
						i, swap_info[i].bdev->bd_dev);
			if (header_data.block_chain[i].size)
				printk("Chain size for device %d is %d.\n", i,
					header_data.block_chain[i].size);
		}
		return -ENOSPC;
	}
	
	if (!currentblockextent) {
		do {
			currentblockchain++;
		} while ((currentblockchain < MAX_SWAPFILES) &&
		   (!header_data.block_chain[currentblockchain].first));

		/* We can validly not have a new blockextent. We
		 * might be compressing data and the user was
		 * too optimistic in setting the compression
		 * ratio or we're just copying the pageset. */

		if (currentblockchain == MAX_SWAPFILES) {
			printk("Argh. Ran out of block chains.\n");
			return -ENOSPC;
		}
				
		currentblockextent = 
		 header_data.block_chain[currentblockchain].first;
		currentblockoffset = currentblockextent->minimum;
		currentblocksperpage = PAGE_SIZE /
		  suspend_bio_ops.get_block_size(swap_info[currentblockchain].bdev);
	}

	submit_params.readahead_index = -1;
	submit_params.page = buffer_page;
	submit_params.dev = swap_info[currentblockchain].bdev;
	submit_params.blocks_used = currentblocksperpage;
		
	/* Get the blocks */
	for (i = 0; i < currentblocksperpage; i++) {
		submit_params.blocks[i] = currentblockoffset;
		GET_EXTENT_NEXT(currentblockextent, currentblockoffset);
	}

	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
		"Writing page %d. Dev is %x. Block is %lu. "
		"Blocksperpage is %d.\n",
		filewriter_page_index,
		submit_params.dev,
		submit_params.blocks[0],
		currentblocksperpage);
		
	suspend_message(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 1,
		"page:%d. bdev:%x. blocks (%d):",
		filewriter_page_index,
		submit_params.dev->bd_dev,
		submit_params.blocks_used);

	for (i = 0; i < currentblocksperpage; i++)
		suspend_message(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 0,
			"0x%lx%s",
			submit_params.blocks[i],
			((i+1) < currentblocksperpage) ? "," : "\n");

	check_shift_keys(0, NULL);

	suspend_bio_ops.submit_io(WRITE, &submit_params, 0);

	filewriter_page_index++;

#endif
	return 0;
}

static int filewriter_write_cleanup(void)
{
#if 0
	if (current_stream == 2) {
		header_data.pd1start_block_extent = currentblockextent;
		header_data.pd1start_block_offset = currentblockoffset;
		header_data.pd1start_chain = currentblockchain;
	}
	
	suspend_bio_ops.finish_all_io();
	
	suspend_bio_ops.check_io_stats();

#endif
	return 0;
}

static int filewriter_read_init(int stream_number)
{
#if 0
	int i;

	if (stream_number == 1) {
		currentblockextent = header_data.pd1start_block_extent;
		currentblockoffset = header_data.pd1start_block_offset;
		currentblockchain = header_data.pd1start_chain;
	} else {
		currentblockextent = NULL;
		currentblockoffset = 0;
		currentblockchain = 0;
		for (currentblockchain = 0; currentblockchain < MAX_SWAPFILES;
			       currentblockchain++)
			if (header_data.block_chain[currentblockchain].first) {
				currentblockextent =
					header_data.block_chain[currentblockchain].first;
				currentblockoffset = currentblockextent->minimum;
				break;
			}

		if (!currentblockextent){
			printk("Error! Can't find any block chain data.\n");
			for (i = 0; i < MAX_SWAPFILES; i++) {
				if (!swap_info[i].swap_file)
					printk("Swap slot %d is unused.\n", i);
				else
					printk("Swap slot %d is device %x.\n",
							i, swap_info[i].bdev->bd_dev);
				if (header_data.block_chain[i].size)
					printk("Chain size for device %d"
							" is %d.\n", i,
					 header_data.block_chain[i].size);
				printk("First entry in chain at %p.\n",
					header_data.block_chain[i].first);
			}
			BUG_ON(1);
		}
	}
	suspend_message(SUSPEND_WRITER, SUSPEND_MEDIUM, 0,
		"Stream %d beginning from position: chain %d. "
			"extent %p, block %ld.\n",
		stream_number,
		currentblockchain, currentblockextent, currentblockoffset);
	
	currentblocksperpage = get_blocks_per_page(currentblockchain);

	filewriter_page_index = 1;

	suspend_bio_ops.reset_io_stats();

	readahead_index = readahead_submit_index = -1;
	readahead_allocs = readahead_frees = 0;

#endif
	return 0;
}

#if 0
static int filewriter_begin_read_chunk(struct page * page, 
		int readahead_index, int sync)
{
	int i;
	struct submit_params submit_params;

	if (currentblockchain == MAX_SWAPFILES) {
		/* Readahead might ask us to read too many blocks */
		printk("Currentblockchain == MAX_SWAPFILES. Begin_read_chunk returning -ENODATA.\n");
		return -ENODATA;
	}

	if (!currentblockextent) {
		do {
			currentblockchain++;
		} while ((!header_data.block_chain[currentblockchain].first) &&
				(currentblockchain < MAX_SWAPFILES));

		/* We can validly not have a new blockextent. We
		 * might have allocated exactly the right amount
		 * of swap for the image and be reading the last
		 * block now.
		 */

		if (currentblockchain == MAX_SWAPFILES) {
			prepare_status(1, 0,
			  "Currentblockchain == MAX_SWAPFILES and "
			  "more data to be read. "
			  "Begin_read_chunk returning -ENOSPC.");
			return -ENOSPC;
		}
				
		currentblockextent =
		  header_data.block_chain[currentblockchain].first;
		currentblockoffset = currentblockextent->minimum;
		currentblocksperpage = get_blocks_per_page(currentblockchain);
	}
	
	submit_params.readahead_index = readahead_index;
	submit_params.page = page;
	submit_params.dev = swap_info[currentblockchain].bdev;
	submit_params.blocks_used = currentblocksperpage;
		
	/* Get the blocks. There is no chance that they span chains. */
	for (i = 0; i < currentblocksperpage; i++) {
		submit_params.blocks[i] = currentblockoffset;
		GET_EXTENT_NEXT(currentblockextent, currentblockoffset);
	}

	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
		"Reading page %d. Dev is %x. Block is %lu. "
		"Blocksperpage is %d. Page is %p(%lx). Readahead index is %d.",
		filewriter_page_index,
		submit_params.dev,
		submit_params.blocks[0],
		currentblocksperpage, 
		page, page_address(page),
		readahead_index);
		
	suspend_message(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 1,
		"page:%d. bdev:%x. blocks (%d):",
		filewriter_page_index,
		submit_params.dev->bd_dev,
		submit_params.blocks_used);

	for (i = 0; i < currentblocksperpage; i++)
		suspend_message(SUSPEND_PAGESETS, SUSPEND_VERBOSE, 0,
			"0x%lx%s",
			submit_params.blocks[i],
			((i+1) < currentblocksperpage) ? "," : "\n");

	check_shift_keys(0, NULL);

	if ((i = suspend_bio_ops.submit_io(READ, &submit_params, sync)))
		return -EPERM;

	filewriter_page_index++;

	check_shift_keys(0, NULL);

	return 0;
}
#endif

/* Note that we ignore the sync parameter. We are implementing
 * read ahead, and will always wait until our readhead buffer has
 * been read before returning.
 */

static int filewriter_read_chunk(struct page * buffer_page, int sync)
{
#if 0
	static int last_result;
	unsigned long * virt;

	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
			"At entrance to filewriter_read_chunk.\n");

	if (sync == SUSPEND_ASYNC)
		return filewriter_begin_read_chunk(buffer_page, -1, sync);

	/* Start new readahead while we wait for our page */
	if (readahead_index == -1) {
		last_result = 0;
		readahead_index = readahead_submit_index = 0;
	}

	/* Start a new readahead? */
	if (last_result) {
		/* We failed to submit a read, and have cleaned up
		 * all the readahead previously submitted */
		if (readahead_submit_index == readahead_index)
			return -EPERM;
		goto wait;
	}
	
	do {
		if (suspend_bio_ops.prepare_readahead(readahead_submit_index))
			break;

		readahead_allocs++;

		suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
			"\nBeginning new readahead %d.\n",
			readahead_submit_index);

		last_result = filewriter_begin_read_chunk(
			suspend_bio_ops.readahead_pages[readahead_submit_index], 
			readahead_submit_index, SUSPEND_ASYNC);
		if (last_result) {
			printk("Begin read chunk for page %d returned %d.\n",
				readahead_submit_index, last_result);
			suspend_bio_ops.cleanup_readahead(readahead_submit_index);
			break;
		}

		readahead_submit_index++;

		if (readahead_submit_index == MAX_READAHEAD)
			readahead_submit_index = 0;

	} while((!last_result) && (readahead_submit_index != readahead_index) &&
			(!suspend_bio_ops.readahead_ready(readahead_index)));

wait:
	suspend_bio_ops.wait_on_readahead(readahead_index);

	virt = kmap_atomic(buffer_page, KM_USER1);
	suspend_message(SUSPEND_WRITER, SUSPEND_HIGH, 0,
		"Returned result of readahead %d,"
		" Copying data from %p to %p.\n", readahead_index,
		page_address(suspend_bio_ops.readahead_pages[readahead_index]),
		virt);

	memcpy(virt, page_address(suspend_bio_ops.readahead_pages[readahead_index]),
			PAGE_SIZE);
	kunmap_atomic(virt, KM_USER1);

	suspend_bio_ops.cleanup_readahead(readahead_index);

	readahead_frees++;

	readahead_index++;
	if (readahead_index == MAX_READAHEAD)
		readahead_index = 0;

#endif
	return 0;
}

static int filewriter_read_cleanup(void)
{
#if 0
	suspend_bio_ops.finish_all_io();
	while (readahead_index != readahead_submit_index) {
		suspend_bio_ops.cleanup_readahead(readahead_index);
		readahead_frees++;
		readahead_index++;
		if (readahead_index == MAX_READAHEAD)
			readahead_index = 0;
	}
	suspend_bio_ops.check_io_stats();
	BUG_ON(readahead_allocs != readahead_frees);
#endif
	return 0;
}

extern unsigned int nr_suspends;

/* filewriter_invalidate_image
 * 
 */
static int filewriter_invalidate_image(void)
{
#if 0
	union p_diskpage cur;
	int result = 0;
	char newsig[11];
	
	cur.address = get_zeroed_page(GFP_ATOMIC);
	if (!cur.address) {
		printk("Unable to allocate a page for restoring the swap signature.\n");
		return -ENOMEM;
	}

	suspend_store_free_mem(SUSPEND_FREE_INVALIDATE_IMAGE, 0);

	/*
	 * If nr_suspends == 0, we must be booting, so no swap pages
	 * will be recorded as used yet.
	 */

	if (nr_suspends > 0)
		filewriter_release_storage();

	/* 
	 * We don't do a sanity check here: we want to restore the swap 
	 * whatever version of kernel made the suspend image.
	 * 
	 * We need to write swap, but swap may not be enabled so
	 * we write the device directly
	 */
	
	suspend_bio_ops.bdev_page_io(READ, resume_block_device,
			resume_firstblock, virt_to_page(cur.pointer));

	result = parse_signature(cur.pointer->swh.magic.magic, 1);
		
	if (result < 4)
		goto out;

	strncpy(newsig, cur.pointer->swh.magic.magic, 10);
	newsig[10] = 0;
	suspend_message(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE, 0,
			"Swap signature will be set to %s.\n", newsig);

	suspend_bio_ops.bdev_page_io(WRITE, resume_block_device, resume_firstblock,
			virt_to_page(cur.pointer));

	if (!nr_suspends)
		printk(KERN_WARNING name_suspend "Image invalidated.\n");
out:
	suspend_bio_ops.finish_all_io();
	free_pages(cur.address, 0);
	suspend_store_free_mem(SUSPEND_FREE_INVALIDATE_IMAGE, 1);
#endif
	return 0;
}

/*
 * workspace_size
 *
 * Description:
 * Returns the number of bytes of RAM needed for this
 * code to do its work. (Used when calculating whether
 * we have enough memory to be able to suspend & resume).
 *
 */
static unsigned long filewriter_memory_needed(void)
{
	return 0;
}

/* Print debug info
 *
 * Description:
 */

static int filewriter_print_debug_stats(char * buffer, int size)
{
#if 0
	int len = 0;
	struct sysinfo sysinfo;
	
	if (active_writer != &filewriterops) {
		len = suspend_snprintf(buffer, size, "- Filewriter inactive.\n");
		return len;
	}

	len = suspend_snprintf(buffer, size, "- Filewriter active.\n");

	si_swapinfo(&sysinfo);
	
	len+= suspend_snprintf(buffer+len, size-len, "  Storage available for image: %ld pages.\n",
			sysinfo.freeswap + filewriter_storage_allocated());

	return len;
#endif
	return 0;
}

/*
 * Storage needed
 *
 * Returns amount of space in the swap header required
 * for the filewriter's data. This ignores the links between
 * pages, which we factor in when allocating the space.
 *
 * We ensure the space is allocated, but actually save the
 * data from write_header_init and therefore don't also define a
 * save_config_info routine.
 */
static unsigned long filewriter_storage_needed(void)
{
#if 0
	return sizeof(header_data);
#endif
	return 0;
}

/*
 * Image_exists
 *
 */

static int filewriter_image_exists(void)
{
#if 0
	int signature_found;
	union p_diskpage diskpage;
	
	if (!resume_device) {
		printk("Not even trying to read header "
				"because resume_device is not set.\n");
		return 0;
	}
	
	//PRINTFREEMEM("at start of filewriter_image_exists.");

	diskpage.address = get_zeroed_page(GFP_ATOMIC);

	/* FIXME: Make sure bdev_page_io handles wrong parameters */
	suspend_bio_ops.bdev_page_io(READ, resume_block_device,
			resume_firstblock, virt_to_page(diskpage.ptr));
	suspend_bio_ops.finish_all_io();
	signature_found = parse_signature(diskpage.pointer->swh.magic.magic, 0);

	if (signature_found < 2) {
		printk(KERN_ERR name_suspend "This is normal swap space.\n" );
		return 0;	/* non fatal error */
	} else if (signature_found == -1) {
		printk(KERN_ERR name_suspend
			"Unable to find a signature. Could you have moved "
			"a swap file?\n");
		return 0;
	} else if (signature_found < 6) {
		if ((!(test_suspend_state(SUSPEND_NORESUME_SPECIFIED)))
				&& suspend_early_boot_message(1,
				SUSPEND_CONTINUE_REQ,
				"Detected the signature of an alternate "
				"implementation.\n"))
			set_suspend_state(SUSPEND_NORESUME_SPECIFIED);
		return 0;
	} else if ((signature_found >> 1) != SIGNATURE_VER) {
		if ((!(test_suspend_state(SUSPEND_NORESUME_SPECIFIED))) &&
			suspend_early_boot_message(1, SUSPEND_CONTINUE_REQ,
			 "Found a different style suspend image signature."))
			set_suspend_state(SUSPEND_NORESUME_SPECIFIED);
	}

	return 1;
#endif
	return 0;
}

/*
 * Mark resume attempted.
 *
 * Record that we tried to resume from this image.
 */

static void filewriter_mark_resume_attempted(void)
{
#if 0
	union p_diskpage diskpage;
	int signature_found;
	
	if (!resume_device) {
		printk("Not even trying to record attempt at resuming"
				" because resume_device is not set.\n");
		return;
	}
	
	diskpage.address = get_zeroed_page(GFP_ATOMIC);

	/* FIXME: Make sure bdev_page_io handles wrong parameters */
	suspend_bio_ops.bdev_page_io(READ, resume_block_device, resume_firstblock, virt_to_page(diskpage.ptr));
	signature_found = parse_signature(diskpage.pointer->swh.magic.magic, 0);

	switch (signature_found) {
		case 12:
		case 13:
			diskpage.pointer->swh.magic.magic[5] |= 0x80;
			break;
	}
	
	suspend_bio_ops.bdev_page_io(WRITE, resume_block_device, resume_firstblock,
			virt_to_page(diskpage.ptr));
	suspend_bio_ops.finish_all_io();
	free_pages(diskpage.address, 0);
	return;
#endif
}

/*
 * Parse Image Location
 *
 * Attempt to parse a resume2= parameter.
 * Swap Writer accepts:
 * resume2=swap:DEVNAME[:FIRSTBLOCK][@BLOCKSIZE]
 *
 * Where:
 * DEVNAME is convertable to a dev_t by name_to_dev_t
 * FIRSTBLOCK is the location of the first block in the swap file
 * (specifying for a swap partition is nonsensical but not prohibited).
 * BLOCKSIZE is the logical blocksize >= 512 & <= PAGE_SIZE, 
 * mod 512 == 0 of the device.
 * Data is validated by attempting to read a swap header from the
 * location given. Failure will result in filewriter refusing to
 * save an image, and a reboot with correct parameters will be
 * necessary.
 */

static int filewriter_parse_image_location(char * commandline, int only_writer)
{
#if 1
	char *thischar, *devstart = NULL, *colon = NULL, *at_symbol = NULL;
	char * diskpage;
	int signature_found;

	if (strncmp(commandline, "file:", 5)) {
		if (!only_writer) {
			printk(name_suspend "Filewriter: Image location doesn't begin with 'file:'\n");
			return 1;
		}
	} else
		commandline += 5;

	devstart = thischar = commandline;
	while ((*thischar != ':') && ((thischar - commandline) < 250) && (*thischar))
		thischar++;

	if (*thischar == ':') {
		colon = thischar;
		*colon = 0;
		thischar++;
	}

	while ((*thischar != '@') && ((thischar - commandline) < 250) && (*thischar))
		thischar++;

	if (*thischar == '@') {
		at_symbol = thischar;
		*at_symbol = 0;
	}
	
	if (colon)
		resume_firstblock = (int) simple_strtoul(colon + 1, NULL, 0);
	else
		resume_firstblock = 0;

	printk("Looking for first block of file header at block %x.\n", resume_firstblock);

	if (at_symbol) {
		resume_firstblocksize = (int) simple_strtoul(at_symbol + 1, NULL, 0);
		if (resume_firstblocksize & 0x1FF)
			printk("Blocksizes are usually a multiple of 512. Don't expect this to work!\n");
	} else
		resume_firstblocksize = 4096;
	printk("Setting logical block size of resume device to %d.\n", resume_firstblocksize);
	
	if (try_to_parse_resume_device(devstart))
		goto invalid;

	if (colon)
		*colon = ':';
	if (at_symbol)
		*at_symbol = '@';

	if ((suspend_bio_ops.get_block_size(resume_block_device) 
				!= resume_firstblocksize) &&
	    (suspend_bio_ops.set_block_size(resume_block_device, resume_firstblocksize)
	    			 == -EINVAL))
		goto invalid;

	diskpage = (char *) get_zeroed_page(GFP_ATOMIC);
	if (suspend_bio_ops.bdev_page_io(READ, resume_block_device, resume_firstblock, virt_to_page(diskpage))) {
		printk(KERN_ERR name_suspend "Failed to submit I/O.\n");
		return -EINVAL;
	}
	suspend_bio_ops.finish_all_io();
	signature_found = parse_signature(diskpage, 0);
	free_page((unsigned long) diskpage);

	if (signature_found != -1) {
		printk(KERN_ERR name_suspend "File signature found.\n");
		return 0;
	}

	printk(KERN_ERR name_suspend "Sorry. No signature found at specified location.\n");
	return -EINVAL;

invalid:
	if (colon)
		*colon = ':';
	if (at_symbol)
		*at_symbol = '@';
	printk(KERN_ERR name_suspend "Sorry. Location looks invalid.\n");
	return -EINVAL;
#endif
}

static int target_is_usable(struct inode * inode)
{

	/* Regular files and block devices are straight forward.
	 * Character devices: Well, suspending to a serial/parallel
	 * port might be handy for preparing an embedded image.
	 * Sockets and FIFOs might be valid in a networking context?
	 *
	 * Directories and links are definitely out.
	 */

	if ((S_ISLNK(inode->i_mode)) ||
	    (S_ISDIR(inode->i_mode)))
		return 0;

	return 1;
}

static void translate_file_to_dev_and_block(void)
{
	mm_segment_t oldfs;
	int fd;
	struct inode * inode;
	char * buffer = (char *) get_zeroed_page(GFP_ATOMIC);
	struct block_device * bdev;
	int offset = 0;

	oldfs = get_fs();
	set_fs(KERNEL_DS);

	BUG_ON(!translate_target);

	fd = sys_open(translate_target, O_RDONLY, 0);

	if (fd < 0) {
		printk("Unable to open %s.\n", translate_target);
		goto cleanup1;
	}

	inode = current->files->fd[fd]->f_dentry->d_inode;

	if (!target_is_usable(inode)) {
		printk("%s is a link or directory. You can't suspend to them!\n",
				translate_target);
		goto cleanup2;
	}

	offset += snprintf(buffer, PAGE_SIZE, "For %s, use resume2=file:",
					translate_target);

	bdev = inode->i_bdev ? inode->i_bdev : inode->i_sb->s_bdev;
			
	if (bdev) {
		char * buffer2 = (char *) get_zeroed_page(GFP_ATOMIC);
		int sector = bmap(inode, 0);

		bdevname(bdev, buffer2);
		offset += snprintf(buffer + offset, PAGE_SIZE - offset, 
				"/dev/%s", buffer2);
		
		if (sector)
			offset += snprintf(buffer + offset, PAGE_SIZE - offset,
				":%x", sector);

		if (inode->i_blksize != PAGE_SIZE)
			offset += snprintf(buffer + offset, PAGE_SIZE - offset,
				"@%lu", inode->i_blksize);
		
		free_pages((unsigned long) buffer2, 0);
	} else
		offset += snprintf(buffer + offset, PAGE_SIZE - offset,
				"%s", translate_target);
			
	offset += snprintf(buffer + offset, PAGE_SIZE - offset,
			"\n");
	printk(buffer);

cleanup2:
	sys_close(fd);

cleanup1:
	set_fs(oldfs);

	free_pages((unsigned long) buffer, 0);
}

extern void attempt_to_parse_resume_device(void);

static struct suspend_proc_data filewriter_proc_data[] = {

	/* Keep me first or change the #define translate_target at the top */
	{
	 .filename			= "filewriter_resume2",
	 .permissions			= PROC_WRITEONLY,
	 .type				= SUSPEND_PROC_DATA_STRING,
	 /* String variable is only allocated when proc used */
	 .write_proc			= translate_file_to_dev_and_block,
	},

	{ .filename			= "disable_filewriter",
	  .permissions			= PROC_RW,
	  .type				= SUSPEND_PROC_DATA_INTEGER,
	  .data = {
		.integer = {
			.variable	= &filewriterops.disabled,
			.minimum	= 0,
			.maximum	= 1,
		}
	  },
	  .write_proc			= attempt_to_parse_resume_device,
	}
};

static struct suspend_plugin_ops filewriterops = {
	.type					= WRITER_PLUGIN,
	.name					= "File Writer",
	.memory_needed				= filewriter_memory_needed,
	.print_debug_info			= filewriter_print_debug_stats,
	.storage_needed				= filewriter_storage_needed,
	.initialise				= filewriter_initialise,
	.cleanup				= filewriter_cleanup,

	.write_init				= filewriter_write_init,
	.write_cleanup				= filewriter_write_cleanup,
	.read_init				= filewriter_read_init,
	.read_cleanup				= filewriter_read_cleanup,

	.ops = {
		.writer = {
		 .write_chunk		= filewriter_write_chunk,
		 .read_chunk		= filewriter_read_chunk,
		 .noresume_reset	= filewriter_noresume_reset,
		 .storage_available 	= filewriter_storage_available,
		 .storage_allocated	= filewriter_storage_allocated,
		 .release_storage	= filewriter_release_storage,
		 .allocate_header_space	= filewriter_allocate_header_space,
		 .allocate_storage	= filewriter_allocate_storage,
		 .image_exists		= filewriter_image_exists,
		 .mark_resume_attempted	= filewriter_mark_resume_attempted,
		 .write_header_init	= filewriter_write_header_init,
		 .write_header_chunk	= filewriter_write_header_chunk,
		 .write_header_cleanup	= filewriter_write_header_cleanup,
		 .read_header_init	= filewriter_read_header_init,
		 .read_header_chunk	= filewriter_read_header_chunk,
		 .read_header_cleanup	= filewriter_read_header_cleanup,
		 .prepare_save_extents	= filewriter_prepare_save_extents,
		 .post_load_extents	= filewriter_post_load_extents,
		 .invalidate_image	= filewriter_invalidate_image,
		 .parse_image_location	= filewriter_parse_image_location,
		}
	}
};

/* ---- Registration ---- */
static __init int filewriter_load(void)
{
	int result;
	int i, numfiles = sizeof(filewriter_proc_data) / sizeof(struct suspend_proc_data);
	
	if (!(result = suspend_register_plugin(&filewriterops))) {
		printk("Software Suspend FleWriter registered.\n");

		for (i=0; i< numfiles; i++)
			suspend_register_procfile(&filewriter_proc_data[i]);
	}
	return result;
}

#ifdef MODULE
static __exit void filewriter_unload(void)
{
	int i, numfiles = sizeof(filewriter_proc_data) / sizeof(struct suspend_proc_data);

	printk("Software Suspend FileWriter unloading.\n");

	for (i=0; i< numfiles; i++)
		suspend_unregister_procfile(&filewriter_proc_data[i]);
	suspend_unregister_plugin(&filewriterops);
}

module_init(filewriter_load);
module_exit(filewriter_unload);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Nigel Cunningham");
MODULE_DESCRIPTION("Suspend2 filewriter");
#else
late_initcall(filewriter_load);
#endif
