/*
 * kernel/power/suspend_userui.c
 *
 * Copyright (C) 2002-2004 Nigel Cunningham <ncunningham@linuxmail.org>
 * Copyright (C) 2005 Bernard Blackham <bernard@blackham.com.au>
 *
 * This file is released under the GPLv2.
 *
 * Routines for Software Suspend's userland UI.
 * 
 */
#define SUSPEND_USERUI_C

#include <linux/suspend.h>
#include <linux/kmod.h>
#include <linux/netlink.h>
#include <linux/security.h>
#include <net/sock.h>
 
#include "plugins.h"
#include "proc.h"
#include "suspend.h"
#include "suspend2_core/suspend.h"
#include "suspend2_core/version.h"
#include "suspend_userui.h"

#include "power.h"

#define NUM_USERUI_SKBS 16 /* Number of preallocated skbs for emergencies */

// #define HANDLE_KEYPRESSES_IN_KERNEL

static char userui_program[256] = "";

/* Number of distinct progress amounts that userspace can display */
static int progress_granularity = 1024;

static struct sock *userui_nl;
static u32 userui_sock_seq;
static unsigned int last_key;
static struct sk_buff *userui_skbs[NUM_USERUI_SKBS];
static int next_free_userui_skb;
static pid_t userui_pid;

DECLARE_WAIT_QUEUE_HEAD(userui_wait_for_process);

static struct sk_buff * userui_get_skb(void)
{
	/* Return a single skbuff either from our pool, or try to allocate one if
	 * the pool is exhausted
	 */
	struct sk_buff *skb;
	if (next_free_userui_skb == -1)
		skb = alloc_skb(NLMSG_SPACE(sizeof(struct userui_msg_params)), GFP_ATOMIC);
	else {
		skb = userui_skbs[next_free_userui_skb];
		userui_skbs[next_free_userui_skb] = NULL;
		next_free_userui_skb--;
	}

	while (next_free_userui_skb < NUM_USERUI_SKBS-1) {
		userui_skbs[next_free_userui_skb+1] =
			alloc_skb(NLMSG_SPACE(sizeof(struct userui_msg_params)), GFP_ATOMIC);
		if (userui_skbs[next_free_userui_skb+1])
			next_free_userui_skb++;
		else
			break;
	}

	return skb;
}

static void userui_send_message(int type, void* params, size_t len)
{
	struct sk_buff *skb;
	struct nlmsghdr *nlh;
	void *dest;

	skb = userui_get_skb();
	if (!skb) {
		printk("suspend_userui: Can't allocate skb!\n");
		return;
	}

	/* NLMSG_PUT contains a hidden goto nlmsg_failure */
	nlh = NLMSG_PUT(skb, 0, userui_sock_seq, type, len);
	userui_sock_seq++;

	dest = NLMSG_DATA(nlh);
	memcpy(dest, params, len);

	netlink_unicast(userui_nl, skb, userui_pid, 0);

	if (!in_atomic() && !irqs_disabled()) {
		struct task_struct *t;
		read_lock(&tasklist_lock);
		if ((t = find_task_by_pid(userui_pid)))
			wake_up_process(t);
		read_unlock(&tasklist_lock);
	}

	return;

nlmsg_failure:
	if (skb)
		kfree_skb(skb);
}

static int userui_nl_set_nofreeze(int pid)
{
	struct task_struct *t;
	userui_pid = pid;

	read_lock(&tasklist_lock);
	if ((t = find_task_by_pid(userui_pid)) == NULL) {
		read_unlock(&tasklist_lock);
		return -EINVAL;
	}

	/* Make sure our userspace UI program can run whilst suspending */
	t->flags |= PF_NOFREEZE;

	read_unlock(&tasklist_lock);

	userui_send_message(USERUI_MSG_NOFREEZE_ACK, NULL, 0);

	return 0;
}

static int userui_nl_ready(int version)
{
	int ret = 0;
	if (version != SUSPEND_USERUI_INTERFACE_VERSION) {
		printk("suspend_userui: Userspace process using invalid interface version\n");
		SET_RESULT_STATE(SUSPEND_ABORTED);
		ret = -EINVAL;
	}

	wake_up_interruptible(&userui_wait_for_process);

	return ret;
}

static int userui_nl_set_state(int n)
{
	/* Only let them change certain settings */
	static const int suspend_action_mask =
		(1 << SUSPEND_REBOOT) | (1 << SUSPEND_PAUSE) | (1 << SUSPEND_SLOW) |
		(1 << SUSPEND_LOGALL) | (1 << SUSPEND_SINGLESTEP) |
		(1 << SUSPEND_PAUSE_NEAR_PAGESET_END);

	suspend_action = (suspend_action & (~suspend_action_mask)) |
		(n & suspend_action_mask);

	return 0;
}

static int userui_nl_set_loglevel(int n)
{
	if (n < 0) n = 0;
	if (n > 9) n = 9;
	console_loglevel = n;
	return 0;
}

static int userui_nl_set_progress_granularity(int n)
{
	if (n < 1) n = 1;
	progress_granularity = n;
	return 0;
}

static int userui_user_rcv_msg(struct sk_buff *skb, struct nlmsghdr *nlh, int *errp)
{
	int type;
	int *data;

	*errp = 0;

	if (!(nlh->nlmsg_flags & NLM_F_REQUEST))
		return 0;

	type = nlh->nlmsg_type;

	/* A control message: ignore them */
	if (type < USERUI_MSG_BASE)
		return 0;

	/* Unknown message: reply with EINVAL */
	if (type >= USERUI_MSG_MAX) {
		*errp = -EINVAL;
		return -1;
	}

	/* All operations require privileges, even GET */
	if (security_netlink_recv(skb)) {
		*errp = -EPERM;
		return -1;
	}

	if (type == USERUI_MSG_NOFREEZE_ME && userui_pid != -1) {
		*errp = -EBUSY;
		return -1;
	}

	data = (int*)NLMSG_DATA(nlh);

	switch (type) {
		case USERUI_MSG_NOFREEZE_ME:
			if ((*errp = userui_nl_set_nofreeze(nlh->nlmsg_pid)))
				return -1;
			break;
		case USERUI_MSG_READY:
			if (nlh->nlmsg_len < NLMSG_LENGTH(sizeof(int))) {
				*errp = -EINVAL;
				return -1;
			}
			if ((*errp = userui_nl_ready(*data)))
				return -1;
			break;
		case USERUI_MSG_ABORT:
			request_abort_suspend();
			break;
		case USERUI_MSG_GET_STATE:
			userui_send_message(USERUI_MSG_GET_STATE, &suspend_action,
					sizeof(suspend_action));
			break;
		case USERUI_MSG_GET_LOGLEVEL:
			userui_send_message(USERUI_MSG_GET_LOGLEVEL, &console_loglevel,
					sizeof(console_loglevel));
			break;
		case USERUI_MSG_SET_STATE:
			if (nlh->nlmsg_len < NLMSG_LENGTH(sizeof(int))) {
				*errp = -EINVAL;
				return -1;
			}
			if ((*errp = userui_nl_set_state(*data)))
				return -1;
			break;
		case USERUI_MSG_SET_LOGLEVEL:
			if (nlh->nlmsg_len < NLMSG_LENGTH(sizeof(int))) {
				*errp = -EINVAL;
				return -1;
			}
			if ((*errp = userui_nl_set_loglevel(*data)))
				return -1;
			break;
		case USERUI_MSG_SET_PROGRESS_GRANULARITY:
			if (nlh->nlmsg_len < NLMSG_LENGTH(sizeof(int))) {
				*errp = -EINVAL;
				return -1;
			}
			if ((*errp = userui_nl_set_progress_granularity(*data)))
				return -1;
			break;
	}

	return 0;
}

static int userui_user_rcv_skb(struct sk_buff *skb)
{
	int err;
	struct nlmsghdr *nlh;

	while (skb->len >= NLMSG_SPACE(0)) {
		u32 rlen;

		nlh = (struct nlmsghdr *) skb->data;
		if (nlh->nlmsg_len < sizeof(*nlh) ||
		    skb->len < nlh->nlmsg_len)
			return 0;
		rlen = NLMSG_ALIGN(nlh->nlmsg_len);
		if (rlen > skb->len)
			rlen = skb->len;
		if (userui_user_rcv_msg(skb, nlh, &err) < 0) {
			if (err == 0)
				return -1;
			netlink_ack(skb, nlh, err);
		} else if (nlh->nlmsg_flags & NLM_F_ACK)
			netlink_ack(skb, nlh, 0);
		skb_pull(skb, rlen);
	}

	return 0;
}

static void userui_netlink_input(struct sock *sk, int len)
{
	do {
		struct sk_buff *skb;
		while ((skb = skb_dequeue(&sk->sk_receive_queue)) != NULL) {
			if (userui_user_rcv_skb(skb)) {
				if (skb->len)
					skb_queue_head(&sk->sk_receive_queue, skb);
				else
					kfree_skb(skb);
				break;
			}
			kfree_skb(skb);
		}
	} while (userui_nl && userui_nl->sk_receive_queue.qlen);
}

static int open_userui_netlink(void)
{
	int i;

	userui_sock_seq = 0x42c0ffee;
	userui_nl = netlink_kernel_create(NETLINK_SUSPEND2_USERUI, userui_netlink_input);
	if (!userui_nl) {
		printk("suspend_userui: Failed to allocate netlink socket.\n");
		return -ENOMEM;
	}

	for (i = 0; i < NUM_USERUI_SKBS; i++)
		userui_skbs[i] = NULL;

	for (i = 0; i < NUM_USERUI_SKBS; i++) {
		userui_skbs[i] = alloc_skb(NLMSG_SPACE(sizeof(struct userui_msg_params)), 0);
		if (userui_skbs[i] == NULL)
			break;
	}
	next_free_userui_skb = i-1;

	return 0;
}

static void close_userui_netlink(void)
{
	int i;

	if (userui_nl) {
		sock_release(userui_nl->sk_socket);
		userui_nl = NULL;
	}

	for (i = 0; i < NUM_USERUI_SKBS; i++)
		if (userui_skbs[i])
			kfree_skb(userui_skbs[i]);
}

static int launch_userui_program(void)
{
	int retval;
	static char *envp[] = {
			"HOME=/",
			"TERM=linux",
			"PATH=/sbin:/usr/sbin:/bin:/usr/bin",
			NULL };
	static char *argv[] = { userui_program, NULL };

	retval = call_usermodehelper(userui_program, argv, envp, 0);

	if (retval)
		printk("suspend_userui: Failed to launch userui program: Error %d\n", retval);

	return retval;
}

static int wait_for_userui_contact(void) 
{
	DECLARE_WAITQUEUE(wait, current);

	add_wait_queue(&userui_wait_for_process, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	/* Wait 10 seconds for the userspace UI process to make contact */
	if (userui_pid == -1 && !TEST_RESULT_STATE(SUSPEND_ABORTED))
		interruptible_sleep_on_timeout(&userui_wait_for_process, 10*HZ);

	set_current_state(TASK_RUNNING);
	remove_wait_queue(&userui_wait_for_process, &wait);

	return 0;
}

static void userui_loglevel_change(void)
{
	userui_send_message(USERUI_MSG_LOGLEVEL_CHANGE, &console_loglevel,
			sizeof(console_loglevel));
}

static void userui_redraw(void)
{
	userui_send_message(USERUI_MSG_REDRAW, NULL, 0);
}

static void userui_send_keypress(void *data)
{
	userui_send_message(USERUI_MSG_KEYPRESS, data, sizeof(unsigned int));
}

DECLARE_WORK(userui_send_keypress_work, userui_send_keypress, &last_key);

static int userui_keypress(unsigned int key)
{
#ifdef HANDLE_KEYPRESSES_IN_KERNEL
	switch (key) {
		case 48:
			console_loglevel = 0;
			break;
		case 49:
			console_loglevel = 1;
			break;
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
		case 122:
			/* `: Toggle slow */
			suspend_action ^= (1 << SUSPEND_SLOW);
			suspend2_core_ops->schedule_message(7);
			break;
		case 1:
			/* F1: Toggle any section debugging. */
			suspend_debug_state ^= (1 << SUSPEND_ANY_SECTION);
			suspend2_core_ops->schedule_message(20);
			break;
		case 2:
			/* F2: Freeze. */
			suspend_debug_state ^= (1 << SUSPEND_FREEZER);
			suspend2_core_ops->schedule_message(21);
			break;
		case 3:
			/* F3: Eat Memory */
			suspend_debug_state ^= (1 << SUSPEND_EAT_MEMORY);
			suspend2_core_ops->schedule_message(22);
			break;
		case 4:
			/* F4: Pagesets. */
			suspend_debug_state ^= (1 << SUSPEND_PAGESETS);
			suspend2_core_ops->schedule_message(23);
			break;
		case 5:
			/* F5: IO. */
			suspend_debug_state ^= (1 << SUSPEND_IO);
			suspend2_core_ops->schedule_message(24);
			break;
		case 6:
			/* F6: Bmapping of pages */
			suspend_debug_state ^= (1 << SUSPEND_BMAP);
			suspend2_core_ops->schedule_message(25);
			break;
		case 7:
			/* F7: Writer */
			suspend_debug_state ^= (1 << SUSPEND_WRITER);
			suspend2_core_ops->schedule_message(26);
			break;
		case 8:
			/* F8: Memory */
			suspend_debug_state ^= (1 << SUSPEND_MEMORY);
			suspend2_core_ops->schedule_message(27);
			break;
		case 9:
			/* F9: Extents */
			suspend_debug_state ^= (1 << SUSPEND_EXTENTS);
			suspend2_core_ops->schedule_message(28);
			break;
		case 10:
			/* F10: Memory Pool */
			suspend_debug_state ^= (1 << SUSPEND_MEM_POOL);
			suspend2_core_ops->schedule_message(29);
			break;
		case 11:
			/* F11: Nosave */
			suspend_debug_state ^= (1 << SUSPEND_NOSAVE);
			suspend2_core_ops->schedule_message(30);
			break;
		case 12:
			/* F12: Integrity */
			suspend_debug_state ^= (1 << SUSPEND_INTEGRITY);
			suspend2_core_ops->schedule_message(31);
			break;
		case 112:
			/* During suspend, toggle pausing with P */
			suspend_action ^= (1 << SUSPEND_PAUSE);
			suspend2_core_ops->schedule_message(1);
			break;
		case 115:
			/* Otherwise, if S pressed, toggle single step */
			suspend_action ^= (1 << SUSPEND_SINGLESTEP);
			suspend2_core_ops->schedule_message(3);
			break;
		case 108:
			/* Otherwise, if L pressed, toggle logging everything */
			suspend_action ^= (1 << SUSPEND_LOGALL);
			suspend2_core_ops->schedule_message(4);
			break;
		case 116:
			/* T: Toggle freezing timers */
			clear_suspend_state(SUSPEND_TIMER_FREEZER_ON);
			suspend2_core_ops->schedule_message(99);
			break;
		case 50:
		case 51:
		case 52:
		case 53:
		case 54:
		case 55:
		case 56:
		case 57:
			console_loglevel = ((key - 48));
			break;
#endif
		default:
			return 0;
	}
	return 1;
#else
	last_key = key;
	schedule_work(&userui_send_keypress_work);
	return 1;
#endif
}

#ifndef HANDLE_KEYPRESSES_IN_KERNEL
static void userui_keypress_excl(int key)
{
	userui_keypress(key);
}
#endif

static void userui_prepare_console(void)
{
	userui_pid = -1;

	if (open_userui_netlink() < 0)
		goto failed;

#ifndef HANDLE_KEYPRESSES_IN_KERNEL
	if (suspend_get_keyboard_exclusive(userui_keypress_excl) < 0)
		goto failed;
#endif

	if (launch_userui_program() < 0)
		goto failed;

	if (wait_for_userui_contact() < 0)
		goto failed;

	if (userui_pid == -1) {
		printk("suspend_userui: Failed to get task for freezerproofing.\n");
		goto failed;
	}

	return;

failed:
	SET_RESULT_STATE(SUSPEND_ABORTED);
}

static void userui_cleanup_console(void)
{
	struct task_struct *t;

	userui_send_message(USERUI_MSG_CLEANUP, NULL, 0);

	read_lock(&tasklist_lock);
	if ((t = find_task_by_pid(userui_pid)))
		t->flags &= ~PF_NOFREEZE;
	read_unlock(&tasklist_lock);

	close_userui_netlink();
	userui_pid = -1;

#ifndef HANDLE_KEYPRESSES_IN_KERNEL
	suspend_release_keyboard_exclusive();
#endif
}

static unsigned long userui_storage_needed(void)
{
	return sizeof(userui_program);
}

static int userui_save_config_info(char *buf)
{
	memcpy(buf, userui_program, sizeof(userui_program));
	return sizeof(userui_program);
}

static void userui_load_config_info(char *buf, int size)
{
	/* Don't load the saved path if one has already been set */
	if (userui_program[0])
		return;

	if (size > sizeof(userui_program))
		size = sizeof(userui_program);

	memcpy(userui_program, buf, size);
	userui_program[sizeof(userui_program)-1] = '\0';
}

static unsigned long userui_memory_needed(void)
{
	/* ball park figure of 128 pages */
	return (128 * PAGE_SIZE);
}

unsigned long userui_update_progress(unsigned long value, unsigned long maximum,
		const char *fmt, va_list args)
{
	static int last_update = -1, last_maximum = 1;
	struct userui_msg_params msg;

	memset(&msg, 0, sizeof(msg));

	msg.a = value;
	msg.b = maximum;

	if (fmt) {
		vsnprintf(msg.text, sizeof(msg.text), fmt, args);
		msg.text[sizeof(msg.text)-1] = '\0';
		last_update = -1; /* force an update */
	}

	if (last_update == -1 ||
			maximum != last_maximum ||
			last_update + (maximum / progress_granularity) < value) {
		userui_send_message(USERUI_MSG_PROGRESS, &msg, sizeof(msg));
		last_update = value;
		last_maximum = maximum;
	}

	return 0;
}

static void userui_message(unsigned long section, unsigned long level,
		int normally_logged,
		const char *fmt, va_list args)
{
	struct userui_msg_params msg;

	memset(&msg, 0, sizeof(msg));

	msg.a = section;
	msg.b = level;
	msg.c = normally_logged;

	if (fmt) {
		vsnprintf(msg.text, sizeof(msg.text), fmt, args);
		msg.text[sizeof(msg.text)-1] = '\0';
	}

	userui_send_message(USERUI_MSG_MESSAGE, &msg, sizeof(msg));
}

/*
 * User interface specific /proc/suspend entries.
 */

static struct suspend_plugin_ops userui_ops;

static struct suspend_proc_data proc_params[] = {
	{ .filename			= "disable_userui_support",
	  .permissions			= PROC_RW,
	  .type				= SUSPEND_PROC_DATA_INTEGER,
	  .data = {
		.integer = {
			.variable	= &userui_ops.disabled,
			.minimum	= 0,
			.maximum	= 1,
		}
	  }
	},
	{ .filename			= "userui_program",
	  .permissions			= PROC_RW,
	  .type				= SUSPEND_PROC_DATA_STRING,
	  .data = {
		.string = {
			.variable	= userui_program,
			.max_length	= 255,
		}
	  }
	}
};

static struct suspend_plugin_ops userui_ops = {
	.type					= UI_PLUGIN,
	.name					= "Userspace UI Support",
	.storage_needed			= userui_storage_needed,
	.save_config_info		= userui_save_config_info,
	.load_config_info		= userui_load_config_info,
	.memory_needed			= userui_memory_needed,
	.ops = {
		.ui = {
			.prepare		= userui_prepare_console,	
			.log_level_change	= userui_loglevel_change,
			.message		= userui_message,
			.update_progress	= userui_update_progress,
			.cleanup		= userui_cleanup_console,
			.keypress		= userui_keypress,
			.post_kernel_restore_redraw =
				userui_redraw,
		}
	}
};

/* ---- Registration ---- */


static __init int userui_mode_load(void)
{
	int i, numfiles = sizeof(proc_params) / sizeof(struct suspend_proc_data);
	int result;

	if (!(result = suspend_register_plugin(&userui_ops))) {
		printk("Software Suspend userspace UI support loaded.\n");
		for (i=0; i< numfiles; i++)
			suspend_register_procfile(&proc_params[i]);
	}

	userui_nl = NULL;
	userui_program[0] = '\0';

	return result;
}

#ifdef MODULE
static __exit void userui_mode_unload(void)
{
	int i, numfiles = sizeof(proc_params) / sizeof(struct suspend_proc_data);

	printk("Software Suspend userspace UI support unloading.\n");

	for (i=0; i< numfiles; i++)
		suspend_unregister_procfile(&proc_params[i]);
	
	suspend_unregister_plugin(&userui_ops);
}

module_init(userui_mode_load);
module_exit(userui_mode_unload);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Bernard Blackham");
MODULE_DESCRIPTION("Suspend2 Userspace UI support");
#else
late_initcall(userui_mode_load);
#endif
