/*
 * lib/dyn_pageflags.c
 *
 * Copyright (C) 2004-2005 Nigel Cunningham <ncunningham@cyclades.com>
 * 
 * This file is released under the GPLv2.
 *
 * Routines for dynamically allocating and releasing bitmaps
 * used as pseudo-pageflags.
 *
 * Arrays are not contiguous. The first sizeof(void *) bytes are
 * the pointer to the next page in the bitmap. This allows us to
 * work under low memory conditions where order 0 might be all
 * that's available. In their original use (suspend2), it also
 * lets us save the pages at suspend time, reload and relocate them
 * as necessary at resume time without much effort.
 */

#include <linux/module.h>
#include <linux/dyn_pageflags.h>

/* clear_map
 *
 * Description:	Clear an array used to store local page flags.
 * Arguments:	dyn_pageflags_t:	The pagemap to be cleared.
 */

void clear_dyn_pageflags(dyn_pageflags_t pagemap)
{
	int i = 0;
	
	for (i = 0; i < PAGES_PER_BITMAP; i++)
		memset((pagemap[i]), 0, PAGE_SIZE);
}

/* allocate_local_pageflags
 *
 * Description:	Allocate a bitmap for local page flags.
 * Arguments:	dyn_pageflags_t *:	Pointer to the bitmap.
 */
int allocate_dyn_pageflags(dyn_pageflags_t *pagemap)
{
	int i;

	BUG_ON(*pagemap);

	*pagemap = kmalloc(sizeof(void *) * PAGES_PER_BITMAP, GFP_ATOMIC);

	for (i = 0; i < PAGES_PER_BITMAP; i++) {
		(*pagemap)[i] = (unsigned long *) get_zeroed_page(GFP_ATOMIC);
		if (!(*pagemap)[i]) {
			printk("Error. Unable to allocate memory for "
					"dynamic pageflags.");
			free_dyn_pageflags(pagemap);
			return 1;
		}
	}
	return 0;
}

/* free_dyn_pageflags
 *
 * Description:	Free a dynamically allocated pageflags bitmap.
 * Arguments:	dyn_pageflags_t *: Pointer to the bitmap being freed.
 */
int free_dyn_pageflags(dyn_pageflags_t *pagemap)
{
	int i = 0;
	if (!*pagemap)
		return 1;
	
	for (i = 0; i < PAGES_PER_BITMAP; i++)
		free_pages((unsigned long) (*pagemap)[i], 0);
	
	kfree(*pagemap);
	*pagemap = NULL;
	return 0;
}

EXPORT_SYMBOL(clear_dyn_pageflags);
EXPORT_SYMBOL(allocate_dyn_pageflags);
EXPORT_SYMBOL(free_dyn_pageflags);
